# This file is Copyright 2020 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
from struct import unpack
from typing import Tuple

try:
    from Crypto.Cipher import ARC4, AES
    from Crypto.Hash import HMAC
except ImportError:
    # Debian/Ubuntu ship pycryptodome under Cryptodome namespace
    from Cryptodome.Cipher import ARC4, AES
    from Cryptodome.Hash import HMAC

from volatility3.framework import interfaces, renderers, exceptions
from volatility3.framework.configuration import requirements
from volatility3.framework.layers import registry
from volatility3.framework.symbols.windows import versions
from volatility3.plugins.windows.registry import hashdump, hivelist, lsadump

vollog = logging.getLogger(__name__)


class Cachedump(interfaces.plugins.PluginInterface):
    """Dumps lsa secrets from memory"""

    _required_framework_version = (2, 0, 0)
    _version = (1, 0, 2)

    @classmethod
    def get_requirements(cls):
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="hivelist", component=hivelist.HiveList, version=(2, 0, 0)
            ),
            requirements.VersionRequirement(
                name="lsadump", component=lsadump.Lsadump, version=(1, 0, 0)
            ),
            requirements.VersionRequirement(
                name="hashdump", component=hashdump.Hashdump, version=(1, 1, 0)
            ),
        ]

    @classmethod
    def get_nlkm(
        cls, sechive: registry.RegistryHive, lsakey: bytes, is_vista_or_later: bool
    ):
        return lsadump.Lsadump.get_secret_by_name(
            sechive, "NL$KM", lsakey, is_vista_or_later
        )

    @classmethod
    def decrypt_hash(cls, edata: bytes, nlkm: bytes, ch, xp: bool):
        if xp:
            hmac_md5 = HMAC.new(nlkm, ch)
            rc4key = hmac_md5.digest()
            rc4 = ARC4.new(rc4key)
            data = rc4.encrypt(edata)  # lgtm [py/weak-cryptographic-algorithm]
        else:
            # Based on code from http://lab.mediaservice.net/code/cachedump.rb
            aes = AES.new(nlkm[16:32], AES.MODE_CBC, ch)
            data = b""
            for i in range(0, len(edata), 16):
                buf = edata[i : i + 16]
                if len(buf) < 16:
                    buf += (16 - len(buf)) * b"\00"
                data += aes.decrypt(buf)
        return data

    @classmethod
    def parse_cache_entry(cls, cache_data: bytes) -> Tuple[int, int, int, bytes, bytes]:
        (uname_len, domain_len) = unpack("<HH", cache_data[:4])
        if len(cache_data[60:62]) == 0:
            return (uname_len, domain_len, 0, b"", b"")
        (domain_name_len,) = unpack("<H", cache_data[60:62])
        ch = cache_data[64:80]
        enc_data = cache_data[96:]
        return (uname_len, domain_len, domain_name_len, enc_data, ch)

    @classmethod
    def parse_decrypted_cache(
        cls, dec_data: bytes, uname_len: int, domain_len: int, domain_name_len: int
    ) -> Tuple[str, str, str, bytes]:
        """Get the data from the cache and separate it into the username, domain name, and hash data"""
        uname_offset = 72
        pad = 2 * ((uname_len / 2) % 2)
        domain_offset = int(uname_offset + uname_len + pad)
        pad = 2 * ((domain_len / 2) % 2)
        domain_name_offset = int(domain_offset + domain_len + pad)
        hashh = dec_data[:0x10]
        username = dec_data[uname_offset : uname_offset + uname_len].decode(
            "utf-16-le", "replace"
        )
        domain = dec_data[domain_offset : domain_offset + domain_len].decode(
            "utf-16-le", "replace"
        )
        domain_name = dec_data[
            domain_name_offset : domain_name_offset + domain_name_len
        ].decode("utf-16-le", "replace")

        return (username, domain, domain_name, hashh)

    def _generator(self, syshive, sechive):
        if not syshive or not sechive:
            if syshive is None:
                vollog.warning("Unable to locate SYSTEM hive")
            if sechive is None:
                vollog.warning("Unable to locate SECURITY hive")
            return None

        bootkey = hashdump.Hashdump.get_bootkey(syshive)
        if not bootkey:
            vollog.warning("Unable to find bootkey")
            return None

        kernel = self.context.modules[self.config["kernel"]]

        vista_or_later = versions.is_vista_or_later(
            context=self.context, symbol_table=kernel.symbol_table_name
        )

        lsakey = lsadump.Lsadump.get_lsa_key(sechive, bootkey, vista_or_later)
        if not lsakey:
            vollog.warning("Unable to find lsa key")
            return None

        nlkm = self.get_nlkm(sechive, lsakey, vista_or_later)
        if not nlkm:
            vollog.warning("Unable to find nlkma key")
            return None

        cache = hashdump.Hashdump.get_hive_key(sechive, "Cache")
        if not cache:
            vollog.warning("Unable to find cache key")
            return None

        for cache_item in cache.get_values():
            if cache_item.Name == "NL$Control":
                continue

            try:
                data = sechive.read(cache_item.Data + 4, cache_item.DataLength)
            except exceptions.InvalidAddressException:
                continue

            if not data:
                continue

            (
                uname_len,
                domain_len,
                domain_name_len,
                enc_data,
                ch,
            ) = self.parse_cache_entry(data)
            # Skip if nothing in this cache entry
            if uname_len == 0 or len(ch) == 0:
                continue
            dec_data = self.decrypt_hash(enc_data, nlkm, ch, not vista_or_later)

            (username, domain, domain_name, hashh) = self.parse_decrypted_cache(
                dec_data, uname_len, domain_len, domain_name_len
            )
            yield (0, (username, domain, domain_name, hashh))

    def run(self):
        offset = self.config.get("offset", None)

        syshive = sechive = None

        for hive in hivelist.HiveList.list_hives(
            context=self.context,
            base_config_path=self.config_path,
            kernel_module_name=self.config["kernel"],
            hive_offsets=None if offset is None else [offset],
        ):
            if hive.get_name().split("\\")[-1].upper() == "SYSTEM":
                syshive = hive
            if hive.get_name().split("\\")[-1].upper() == "SECURITY":
                sechive = hive

        return renderers.TreeGrid(
            [("Username", str), ("Domain", str), ("Domain name", str), ("Hash", bytes)],
            self._generator(syshive, sechive),
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
"""All plugins output a TreeGrid object which must then be rendered (either by a
GUI, or as text output, html output or in some other form.

This module defines both the output format (:class:`TreeGrid`) and the
renderer interface which can interact with a TreeGrid to produce
suitable output.
"""

import datetime
import warnings
from abc import ABCMeta, abstractmethod
from collections import abc
from typing import (
    Any,
    Callable,
    ClassVar,
    Generator,
    List,
    NamedTuple,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)
from typing import Dict

from volatility3.framework import interfaces


class BasicType:
    def __str__(self) -> str:
        """Fallback method for rendering basic types"""
        return str(self)


class BaseAbsentValue:
    """Class that represents values which are not present for some reason."""

    def __str__(self) -> str:
        """Fallback method for rendering basic types"""
        return "-"


class Column(NamedTuple):
    name: str
    type: Any


RenderOption = Any

T = TypeVar("T")


class TypeRendererInterface:
    type = T

    def __init__(
        self, func: Optional[Callable] = None, options: Optional[Dict[str, Any]] = None
    ):
        self._options = options or {}
        setattr(self, "render", func)

    @property
    def options(self):
        return self._options

    def render(self, data: Union[T, BaseAbsentValue]) -> Any:
        """Renders a specific datatype"""
        return ""

    def __call__(self, data: Union[T, BaseAbsentValue]) -> Any:
        """Shortcut for render"""
        return self.render(data)


class Renderer(metaclass=ABCMeta):
    """Class that defines the interface that all output renderers must
    support."""

    _type_renderers: Dict[Union[Type, str], Callable]

    def __init__(self, options: Optional[List[RenderOption]] = None) -> None:
        """Accepts an options object to configure the renderers."""
        # FIXME: Once the config option objects are in place, put the _type_check in place

    @abstractmethod
    def get_render_options(self) -> List[RenderOption]:
        """Returns a list of rendering options."""

    @abstractmethod
    def render(self, grid: "TreeGrid") -> None:
        """Takes a grid object and renders it based on the object's
        preferences."""


class ColumnSortKey(metaclass=ABCMeta):
    ascending: bool = True

    @abstractmethod
    def __call__(self, values: List[Any]) -> Any:
        """The key function passed as a sort key to the TreeGrid's visit
        function."""


class TreeNode(abc.Sequence, metaclass=ABCMeta):
    def __init__(self, path, treegrid, parent, values):
        """Initializes the TreeNode."""

    @property
    @abstractmethod
    def values(self) -> List["BaseTypes"]:
        """Returns the list of values from the particular node, based on column
        index."""

    @property
    @abstractmethod
    def path(self) -> str:
        """Returns a path identifying string.

        This should be seen as opaque by external classes, Parsing of
        path locations based on this string are not guaranteed to remain
        stable.
        """

    @property
    @abstractmethod
    def parent(self) -> Optional["TreeNode"]:
        """Returns the parent node of this node or None."""

    @property
    @abstractmethod
    def path_depth(self) -> int:
        """Return the path depth of the current node."""

    @abstractmethod
    def path_changed(self, path: str, added: bool = False) -> None:
        """Updates the path based on the addition or removal of a node higher
        up in the tree.

        This should only be called by the containing TreeGrid and
        expects to only be called for affected nodes.
        """


class Disassembly(BasicType):
    """A class to indicate that the bytes provided should be disassembled
    (based on the architecture)"""

    possible_architectures = ["intel", "intel64", "arm", "arm64"]

    def __init__(
        self, data: bytes, offset: int = 0, architecture: str = "intel64"
    ) -> None:
        warnings.warn(
            "interfaces.renderers.Disassembly is now renderers.Disassembly",
            FutureWarning,
        )
        self.data = data
        self.architecture = None
        if architecture in self.possible_architectures:
            self.architecture = architecture
        if not isinstance(offset, int):
            raise TypeError("Offset must be an integer type")
        self.offset = offset

    def __str__(self) -> str:
        """Fallback method of rendering"""
        return str(self.data)


# We don't class these off a shared base, because the BaseTypes must only
# contain the types that the validator will accept (which would not include the base)

_Type = TypeVar("_Type")
BaseTypes = Union[
    Type[int],
    Type[str],
    Type[float],
    Type[bytes],
    Type[datetime.datetime],
    Type[BaseAbsentValue],
    Type[BasicType],
]
ColumnsType = List[Tuple[str, BaseTypes]]
VisitorSignature = Callable[[TreeNode, _Type], _Type]


class TreeGrid(metaclass=ABCMeta):
    """Class providing the interface for a TreeGrid (which contains TreeNodes)

    The structure of a TreeGrid is designed to maintain the structure of the tree in a single object.
    For this reason each TreeNode does not hold its children, they are managed by the top level object.
    This leaves the Nodes as simple data carries and prevents them being used to manipulate the tree as a whole.
    This is a data structure, and is not expected to be modified much once created.

    Carrying the children under the parent makes recursion easier, but then every node is its own little tree
    and must have all the supporting tree functions.  It also allows for a node to be present in several different trees,
    and to create cycles.
    """

    # TODO: Figure out why this isn't just BaseTypes (which includes AbsentValues'
    base_types: ClassVar[Tuple] = (int, str, float, bytes, datetime.datetime, BasicType)

    def __init__(
        self,
        columns: ColumnsType,
        generator: Generator,
        context: Optional["interfaces.context.ContextInterface"] = None,
    ) -> None:
        """Constructs a TreeGrid object using a specific set of columns.

        The TreeGrid itself is a root element, that can have children but no values.
        The TreeGrid does *not* contain any information about formatting,
        these are up to the renderers and plugins.

        Args:
            columns: A list of column tuples made up of (name, type).
            generator: An iterable containing row for a tree grid, each row contains a indent level followed by the values for each column in order.
        """
        self._context = context

    @property
    def context(self) -> Optional["interfaces.context.ContextInterface"]:
        """Returns the context value for the tree grid (to retrieve data items)

        This is a property to ensure the renderers don't try changing the context for any reason
        """
        return self._context

    @staticmethod
    @abstractmethod
    def sanitize_name(text: str) -> str:
        """Method used to sanitize column names for TreeNodes."""

    @abstractmethod
    def populate(
        self,
        function: Optional[VisitorSignature] = None,
        initial_accumulator: Any = None,
        fail_on_errors: bool = True,
    ) -> Optional[Exception]:
        """Populates the tree by consuming the TreeGrid's construction
        generator Func is called on every node, so can be used to create output
        on demand.

        This is equivalent to a one-time visit.
        """

    @property
    @abstractmethod
    def populated(self) -> bool:
        """Indicates that population has completed and the tree may now be
        manipulated separately."""

    @property
    @abstractmethod
    def columns(self) -> List[Column]:
        """Returns the available columns and their ordering and types."""

    @abstractmethod
    def children(self, node: TreeNode) -> List[TreeNode]:
        """Returns the subnodes of a particular node in order."""

    @abstractmethod
    def values(self, node: TreeNode) -> Tuple[BaseTypes, ...]:
        """Returns the values for a particular node.

        The values returned are mutable,
        """

    @abstractmethod
    def is_ancestor(self, node: TreeNode, descendant: TreeNode) -> bool:
        """Returns true if descendent is a child, grandchild, etc of node."""

    @abstractmethod
    def max_depth(self) -> int:
        """Returns the maximum depth of the tree."""

    @staticmethod
    def path_depth(node: TreeNode) -> int:
        """Returns the path depth of a particular node."""
        return node.path_depth

    @abstractmethod
    def visit(
        self,
        node: Optional[TreeNode],
        function: VisitorSignature,
        initial_accumulator: _Type,
        sort_key: Optional[ColumnSortKey] = None,
    ) -> None:
        """Visits all the nodes in a tree, calling function on each one.

        function should have the signature function(node, accumulator) and return new_accumulator
        If accumulators are not needed, the function must still accept a second parameter.

        The order of that the nodes are visited is always depth first, however, the order children are traversed can
        be set based on a sort_key function which should accept a node's values and return something that can be
        sorted to receive the desired order (similar to the sort/sorted key).

        If node is None, then the root node is used.

        Args:
            node: The initial node to be visited
            function: The visitor to apply to the nodes under the initial node
            initial_accumulator: An accumulator that allows data to be transferred between one visitor call to the next
            sort_key: Information about the sort order of columns in order to determine the ordering of results
        """



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import datetime
import logging
import sys
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    TextIO,
)
from volatility3.framework import interfaces, renderers
from volatility3.framework.renderers import format_hints
from volatility3.cli import text_renderer

vollog = logging.getLogger(__name__)

ARROW_PRESENT = False
try:
    import pyarrow as pa
    import pyarrow.parquet as pq

    ARROW_PRESENT = True
except ImportError:
    vollog.debug("Arrow/Parquet libraries not found")


class ArrowRenderer(text_renderer.CLIRenderer):
    """Renderer that outputs Arrow IPC format data."""

    name = "arrow"
    structured_output = True
    _version = (1, 0, 0)

    def __init__(
        self, options: Optional[List[interfaces.renderers.RenderOption]] = None
    ) -> None:
        super().__init__(options)

        if not ARROW_PRESENT:
            raise RuntimeError("Arrow output format requires the pyarrow package")

        self._to_arrow_type = {
            renderers.Disassembly: pa.utf8,
            bool: pa.bool_,
            int: pa.int64,
            float: pa.float64,
            str: pa.utf8,
            datetime.datetime: lambda: pa.timestamp("ms"),
            format_hints.Bin: pa.uint64,
            format_hints.Hex: pa.uint64,
            format_hints.MultiTypeData: pa.binary,
            format_hints.HexBytes: pa.binary,
            renderers.LayerData: pa.binary,
            bytes: pa.binary,
        }

        # indicates if the output from the plugin is nested, e.g., pstree
        # which would then need to be flattened
        self._is_tree_result = False
        self._node_id_counter = 0

    def get_render_options(self) -> List[interfaces.renderers.RenderOption]:
        return []

    def to_arrow_schema(self, grid: interfaces.renderers.TreeGrid) -> "pa.Schema":
        fields = []
        for column in grid.columns:
            arrow_type = self._to_arrow_type[column.type]
            fields.append(pa.field(column.name, arrow_type()))

        # if the output is nested, e.g., windows.pstree
        if self._is_tree_result:
            fields.append(pa.field("_vol_id", pa.uint64()))
            fields.append(pa.field("_vol_parent_id", pa.uint64()))

        return pa.schema(fields)

    def _flatten_tree_structure(self, nested: List[Dict]) -> List[Dict]:
        """
        Flattens a list of nested dicts using the `__children` key.

        Each node gets a `_vol_id` and a `_vol_parent_id` to preserve
        the original tree structure in a flat format suitable for tabular output.

        Args:
            nested: A list of dicts with optional `__children` lists (tree nodes).

        Returns:
            A flat list of dicts with `_vol_id` and `_vol_parent_id`.
        """
        rows = []
        self._node_id_counter = 0

        def _process_node(node: Dict, parent_id: Optional[int]):
            current_id = self._node_id_counter
            self._node_id_counter += 1

            entry = {k: v for k, v in node.items() if k != "__children"}
            entry["_vol_id"] = current_id
            entry["_vol_parent_id"] = parent_id
            rows.append(entry)

            for child in node.get("__children", []):
                _process_node(child, current_id)

        for root in nested:
            _process_node(root, None)

        return rows

    def output_result(self, schema: "pa.Schema", outfd: TextIO, result):
        """Outputs the JSON data to a file in a particular format"""

        if self._is_tree_result:
            result = self._flatten_tree_structure(result)

        t = pa.Table.from_pylist(result, schema=schema)
        self.write_table(t, outfd)

    def write_table(self, t: "pa.Table", outfd: TextIO) -> None:
        buf = pa.BufferOutputStream()

        writer = pa.ipc.new_stream(buf, t.schema)
        writer.write_table(t)
        writer.close()

        # Get the buffer bytes and write to output
        buf_bytes = buf.getvalue().to_pybytes()
        outfd.buffer.write(buf_bytes)

    def render(self, grid: interfaces.renderers.TreeGrid):
        outfd = sys.stdout
        final_output: Tuple[
            Dict[str, List[interfaces.renderers.TreeNode]],
            List[interfaces.renderers.TreeNode],
        ] = ({}, [])

        ignore_columns = self.ignored_columns(grid)

        def visitor(
            node: interfaces.renderers.TreeNode,
            accumulator: Tuple[Dict[str, Dict[str, Any]], List[Dict[str, Any]]],
        ) -> Tuple[Dict[str, Dict[str, Any]], List[Dict[str, Any]]]:
            # Nodes always have a path value, giving them a path_depth of at least 1, we use max just in case
            acc_map, final_tree = accumulator
            node_dict: Dict[str, Any] = {"__children": []}
            line = []
            for column_index, column in enumerate(grid.columns):
                if column in ignore_columns:
                    continue

                data = list(node.values)[column_index]

                if isinstance(data, interfaces.renderers.BaseAbsentValue):
                    data = None

                if isinstance(data, renderers.Disassembly):
                    data = text_renderer.display_disassembly(data)

                if isinstance(data, renderers.LayerData):
                    data = text_renderer.LayerDataRenderer().render_bytes(data)[0]

                node_dict[column.name] = data
                line.append(data)

            if self.filter and self.filter.filter(line):
                return accumulator

            if node.parent:
                acc_map[node.parent.path]["__children"].append(node_dict)
                self._is_tree_result = True
            else:
                final_tree.append(node_dict)
            acc_map[node.path] = node_dict

            return (acc_map, final_tree)

        if not grid.populated:
            grid.populate(visitor, final_output)
        else:
            grid.visit(node=None, function=visitor, initial_accumulator=final_output)

        schema = self.to_arrow_schema(grid)
        self.output_result(schema, outfd, final_output[1])


class ParquetRenderer(ArrowRenderer):
    """Renderer that outputs Parquet format data."""

    name = "parquet"
    structured_output = True
    _version = (1, 0, 0)

    def get_render_options(self) -> List[interfaces.renderers.RenderOption]:
        return []

    def write_table(self, table: "pa.Table", outfd: TextIO) -> None:
        """
        Writes a table to stdout using the Parquet format.

        Args:
            t: The Arrow table to write
            outfd: The output file descriptor

        Returns:
            Nothing
        """
        # Write DataFrame to a temporary file-like object
        buf = pa.BufferOutputStream()
        pq.write_table(table, buf, compression="snappy")

        # Get the buffer as a bytes object
        buf_bytes = buf.getvalue().to_pybytes()
        outfd.buffer.write(buf_bytes)



# This file is Copyright 2022 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
import urllib.parse
from typing import Optional, Any, List

try:
    import s3fs

    HAS_S3FS = True
except ImportError:
    HAS_S3FS = False

try:
    import gcsfs

    HAS_GCSFS = True
except ImportError:
    HAS_GCSFS = False

from volatility3.framework.layers import resources

vollog = logging.getLogger(__file__)

if HAS_S3FS:

    class S3FileSystemHandler(resources.VolatilityHandler):
        @classmethod
        def non_cached_schemes(cls) -> List[str]:
            return ["s3"]

        @staticmethod
        def default_open(req: urllib.request.Request) -> Optional[Any]:
            """Handles the request if it's the s3 scheme."""
            if req.type == "s3":
                object_uri = "://".join(req.full_url.split("://")[1:])
                return s3fs.S3FileSystem().open(object_uri)
            return None


if HAS_GCSFS:

    class GSFileSystemHandler(resources.VolatilityHandler):
        @classmethod
        def non_cached_schemes(cls) -> List[str]:
            return ["gs"]

        @staticmethod
        def default_open(req: urllib.request.Request) -> Optional[Any]:
            """Handles the request if it's the gs scheme."""
            if req.type == "gs":
                object_uri = "://".join(req.full_url.split("://")[1:])
                return gcsfs.GCSFileSystem().open(object_uri)
            return None



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import datetime
import logging
from typing import Iterable, List, Optional, Sequence, Tuple, Union

from volatility3.framework import constants, exceptions, interfaces, objects, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.layers import registry as registry_layer
from volatility3.framework.renderers import conversion, format_hints
from volatility3.framework.symbols.windows.extensions import registry
from volatility3.plugins.windows.registry import hivelist

vollog = logging.getLogger(__name__)


class PrintKey(interfaces.plugins.PluginInterface):
    """Lists the registry keys under a hive or specific key value."""

    _required_framework_version = (2, 0, 0)
    _version = (1, 1, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="hivelist", component=hivelist.HiveList, version=(2, 0, 0)
            ),
            requirements.IntRequirement(
                name="offset", description="Hive Offset", default=None, optional=True
            ),
            requirements.StringRequirement(
                name="key", description="Key to start from", default=None, optional=True
            ),
            requirements.BooleanRequirement(
                name="recurse",
                description="Recurses through keys",
                default=False,
                optional=True,
            ),
        ]

    @classmethod
    def key_iterator(
        cls,
        hive: registry_layer.RegistryHive,
        node_path: Optional[Sequence[objects.StructType]] = None,
        recurse: bool = False,
    ) -> Iterable[
        Tuple[
            int, bool, datetime.datetime, str, bool, interfaces.objects.ObjectInterface
        ]
    ]:
        """Walks through a set of nodes from a given node (last one in
        node_path). Avoids loops by not traversing into nodes already present
        in the node_path.

        Args:
            hive: The registry hive to walk
            node_path: The list of nodes that make up the
            recurse: Traverse down the node tree or stay only on the same level

        Yields:
            A tuple of results (depth, is_key, last write time, path, volatile, and the node).
        """
        if not node_path:
            node_path = [hive.get_node(hive.root_cell_offset)]
        if not isinstance(node_path, list) or len(node_path) < 1:
            vollog.warning("Hive walker was not passed a valid node_path (or None)")
            return None
        node = node_path[-1]
        key_path_items = [hive] + node_path[1:]
        key_path_names = []
        for k in key_path_items:
            try:
                key_path_names.append(k.get_name())
            except (
                registry_layer.InvalidAddressException,
                registry_layer.RegistryException,
            ):
                key_path_names.append("-")
        key_path = "\\".join([k for k in key_path_names])

        if node.vol.type_name.endswith(constants.BANG + "_CELL_DATA"):
            raise registry_layer.RegistryFormatException(
                hive.name, "Encountered _CELL_DATA instead of _CM_KEY_NODE"
            )
        last_write_time = conversion.wintime_to_datetime(node.LastWriteTime.QuadPart)

        for key_node in node.get_subkeys():
            result = (
                len(node_path),
                True,
                last_write_time,
                key_path,
                key_node.get_volatile(),
                key_node,
            )
            yield result

            if recurse:
                if key_node.vol.offset not in [x.vol.offset for x in node_path]:
                    try:
                        key_node.get_name()
                    except (
                        exceptions.InvalidAddressException,
                        registry_layer.RegistryException,
                    ) as excp:
                        vollog.debug(excp)
                        continue

                    yield from cls.key_iterator(
                        hive, node_path + [key_node], recurse=recurse
                    )

        for value_node in node.get_values():
            result = (
                len(node_path),
                False,
                last_write_time,
                key_path,
                node.get_volatile(),
                value_node,
            )
            yield result

    def _printkey_iterator(
        self,
        hive: registry_layer.RegistryHive,
        node_path: Optional[Sequence[objects.StructType]] = None,
        recurse: bool = False,
    ):
        """Method that wraps the more generic key_iterator, to provide output
        for printkey specifically.

        Args:
            hive: The registry hive to walk
            node_path: The list of nodes that make up the
            recurse: Traverse down the node tree or stay only on the same level

        Yields:
            The depth, and a tuple of results (last write time, hive offset, type, path, name, data and volatile)
        """
        for (
            depth,
            is_key,
            last_write_time,
            key_path,
            volatile,
            node,
        ) in self.key_iterator(hive, node_path, recurse):
            if is_key:
                try:
                    key_node_name = node.get_name()
                except (
                    exceptions.InvalidAddressException,
                    registry_layer.RegistryException,
                ) as excp:
                    vollog.debug(excp)
                    key_node_name = renderers.UnreadableValue()

                # if the item is a subkey, use the LastWriteTime of that subkey
                last_write_time = conversion.wintime_to_datetime(
                    node.LastWriteTime.QuadPart
                )

                yield (
                    depth,
                    (
                        last_write_time,
                        renderers.format_hints.Hex(hive.hive_offset),
                        "Key",
                        key_path,
                        key_node_name,
                        renderers.NotApplicableValue(),
                        volatile,
                    ),
                )
            else:
                try:
                    value_node_name = node.get_name() or "(Default)"
                except (
                    exceptions.InvalidAddressException,
                    registry_layer.RegistryException,
                ) as excp:
                    vollog.debug(excp)
                    value_node_name = renderers.UnreadableValue()

                try:
                    value_type = registry.RegValueTypes(node.Type).name
                except (
                    exceptions.InvalidAddressException,
                    registry_layer.RegistryException,
                ) as excp:
                    vollog.debug(excp)
                    value_type = renderers.UnreadableValue()

                if isinstance(value_type, renderers.UnreadableValue):
                    vollog.debug(
                        "Couldn't read registry value type, so data is unreadable"
                    )
                    value_data: Union[interfaces.renderers.BaseAbsentValue, bytes] = (
                        renderers.UnreadableValue()
                    )
                else:
                    try:
                        value_data = node.decode_data()

                        if isinstance(value_data, int):
                            value_data = format_hints.MultiTypeData(
                                value_data, encoding="utf-8"
                            )
                        elif (
                            registry.RegValueTypes(node.Type)
                            == registry.RegValueTypes.REG_BINARY
                        ):
                            value_data = format_hints.MultiTypeData(
                                value_data, show_hex=True
                            )
                        elif (
                            registry.RegValueTypes(node.Type)
                            == registry.RegValueTypes.REG_MULTI_SZ
                        ):
                            value_data = format_hints.MultiTypeData(
                                value_data, encoding="utf-16-le", split_nulls=True
                            )
                        else:
                            value_data = format_hints.MultiTypeData(
                                value_data, encoding="utf-16-le"
                            )
                    except (
                        ValueError,
                        exceptions.InvalidAddressException,
                        registry_layer.RegistryException,
                    ) as excp:
                        vollog.debug(excp)
                        value_data = renderers.UnreadableValue()

                result = (
                    depth,
                    (
                        last_write_time,
                        renderers.format_hints.Hex(hive.hive_offset),
                        value_type,
                        key_path,
                        value_node_name,
                        value_data,
                        volatile,
                    ),
                )
                yield result

    def _registry_walker(
        self,
        hive_offsets: Optional[List[int]] = None,
        key: Optional[str] = None,
        recurse: bool = False,
    ):
        for hive in hivelist.HiveList.list_hives(
            context=self.context,
            base_config_path=self.config_path,
            kernel_module_name=self.config["kernel"],
            hive_offsets=hive_offsets,
        ):
            try:
                # Walk it
                if key is not None:
                    node_path = hive.get_key(key, return_list=True)
                else:
                    node_path = [hive.get_node(hive.root_cell_offset)]
                for x, y in self._printkey_iterator(hive, node_path, recurse=recurse):
                    yield (x - len(node_path), y)
            except (
                exceptions.InvalidAddressException,
                KeyError,
                registry_layer.RegistryException,
            ) as excp:
                if isinstance(excp, KeyError):
                    vollog.debug(
                        f"Key '{key}' not found in Hive at offset {hex(hive.hive_offset)}."
                    )
                elif isinstance(excp, registry_layer.RegistryException):
                    vollog.debug(excp)
                elif isinstance(excp, exceptions.InvalidAddressException):
                    vollog.debug(
                        f"Invalid address identified in Hive: {hex(excp.invalid_address)}"
                    )
                result = (
                    0,
                    (
                        renderers.UnreadableValue(),
                        format_hints.Hex(hive.hive_offset),
                        "Key",
                        f"{hive.get_name()}\\" + (key or ""),
                        renderers.UnreadableValue(),
                        renderers.UnreadableValue(),
                        renderers.UnreadableValue(),
                    ),
                )
                yield result

    def run(self):
        offset = self.config.get("offset", None)

        return renderers.TreeGrid(
            columns=[
                ("Last Write Time", datetime.datetime),
                ("Hive Offset", format_hints.Hex),
                ("Type", str),
                ("Key", str),
                ("Name", str),
                ("Data", format_hints.MultiTypeData),
                ("Volatile", bool),
            ],
            generator=self._registry_walker(
                hive_offsets=None if offset is None else [offset],
                key=self.config.get("key", None),
                recurse=self.config.get("recurse", None),
            ),
        )



# This file is Copyright 2022 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
from typing import Dict, Iterator, List, Optional, Tuple

from volatility3.framework import constants, exceptions, interfaces, objects

vollog = logging.getLogger(__name__)


class MFTEntry(objects.StructType):
    """This represents the base MFT Record"""

    def __init__(
        self,
        context: interfaces.context.ContextInterface,
        type_name: str,
        object_info: interfaces.objects.ObjectInformation,
        size: int,
        members: Dict[str, Tuple[int, interfaces.objects.Template]],
    ) -> None:
        super().__init__(context, type_name, object_info, size, members)

        self._attrs_loaded = False
        self._attrs: List[MFTAttribute] = []

    @property
    def symbol_table_name(self) -> str:
        return self.vol.type_name.split(constants.BANG)[0]

    def get_signature(self) -> objects.String:
        signature = self.Signature.cast("string", max_length=4, encoding="latin-1")
        return signature

    @property
    def attributes(self) -> Iterator["MFTAttribute"]:
        """
        Lazily evaluate and yield attributes, caching them in an internal list
        for re-retrieval.
        """
        if not self._attrs_loaded:
            self._attrs = list(self._attributes())
            self._attrs_loaded = True

        yield from self._attrs

    def longest_filename(self) -> Optional[objects.String]:
        names = [name.get_full_name() for name in self.filename_entries()]
        if not names:
            return None

        return max(names, key=lambda x: len(str(x)))

    def _attributes(self) -> Iterator["MFTAttribute"]:
        # We will update this on each pass in the next loop and use it as the new offset.
        attr_base_offset = self.FirstAttrOffset
        attribute_object_type_name = (
            self.symbol_table_name + constants.BANG + "ATTRIBUTE"
        )

        attr: MFTAttribute = self._context.object(
            attribute_object_type_name,
            offset=self.vol.offset + attr_base_offset,
            layer_name=self.vol.layer_name,
        )

        # There is no field that has a count of Attributes
        # Keep Attempting to read attributes until we get an invalid attr_header.AttrType
        try:
            while attr.Attr_Header.AttrType.is_valid_choice:
                yield attr

                # If there's no advancement the loop will never end, so break it now
                if attr.Attr_Header.Length == 0:
                    break

                # Update the base offset to point to the next attribute
                attr_base_offset += attr.Attr_Header.Length
                # Get the next attribute
                attr: MFTAttribute = self._context.object(
                    attribute_object_type_name,
                    offset=self.vol.offset + attr_base_offset,
                    layer_name=self.vol.layer_name,
                )
        except exceptions.InvalidAddressException as e:
            vollog.debug(
                f"Failed to read attribute at {attr.vol.offset:#x}: {e.__class__.__name__}"
            )
            return

    def standard_information_entries(
        self,
    ) -> Iterator[objects.StructType]:
        """
        Yields a STANDARD_INFORMATION struct for each of the
        STANDARD_INFORMATION attributes in this MFT record (although there
        should only be one per record).
        """
        for attr in self.attributes:
            attr_type = attr.Attr_Header.AttrType.lookup()
            if attr_type != "STANDARD_INFORMATION":
                continue

            si_object = (
                self.symbol_table_name + constants.BANG + "STANDARD_INFORMATION_ENTRY"
            )

            yield attr.Attr_Data.cast(si_object)

    def filename_entries(self) -> Iterator["MFTFileName"]:
        """
        Yields an MFT Filename for each of the FILE_NAME attributes contained
        in this MFT record. There are often two - one for the long filename,
        and the other with the DOS 8.3 short name.
        """
        for attr in self.attributes:
            try:
                attr_type = attr.Attr_Header.AttrType.lookup()
                if attr_type != "FILE_NAME":
                    continue

                fn_object = self.symbol_table_name + constants.BANG + "FILE_NAME_ENTRY"
                attr_data = attr.Attr_Data.cast(fn_object)
            except exceptions.InvalidAddressException as e:
                vollog.debug(
                    f"Failed to read attr at {attr.vol.offset:#x}: {e.__class__.__name__}"
                )
                continue
            yield attr_data

    def _data_attributes(self):
        for attr in self.attributes:
            if not (
                attr.Attr_Header.AttrType.lookup() == "DATA"
                and attr.Attr_Header.NonResidentFlag == 0
            ):
                continue

            yield attr

    def resident_data_attributes(self) -> Iterator["MFTAttribute"]:
        """
        Yields all MFT attributes that contain resident data for the primary
        stream.
        """
        for attr in self._data_attributes():
            if attr.Attr_Header.NameLength == 0:
                yield attr

    def alternate_data_streams(self) -> Iterator["MFTAttribute"]:
        """
        Yields all MFT attributes that contain alternate data streams (ADS).
        """
        for attr in self._data_attributes():
            if attr.Attr_Header.NameLength != 0:
                yield attr


class MFTFileName(objects.StructType):
    """This represents an MFT $FILE_NAME Attribute"""

    def get_full_name(self) -> objects.String:
        """
        Returns the UTF-16 decoded filename.
        """
        output = self.Name.cast(
            "string", encoding="utf16", max_length=self.NameLength * 2, errors="replace"
        )
        return output


class MFTAttribute(objects.StructType):
    """This represents an MFT ATTRIBUTE"""

    def get_resident_filename(self) -> Optional[objects.String]:
        """
        Returns the resident filename (typically for an Alternate Data Stream (ADS)).
        """
        # 4MB chosen as cutoff instead of 4KB to allow for recovery from format /L created file systems
        # Length as 512 as its 256*2, which is the maximum size for an entire file path, so this is even generous
        if (
            self.Attr_Header.ContentOffset > 0x400000
            or self.Attr_Header.NameLength > 512
        ):
            return None

        # To get the resident name, we jump to relative name offset and read name length * 2 bytes of data
        try:
            name = self._context.object(
                self.vol.type_name.split(constants.BANG)[0] + constants.BANG + "string",
                layer_name=self.vol.layer_name,
                offset=self.vol.offset + self.Attr_Header.NameOffset,
                max_length=self.Attr_Header.NameLength * 2,
                errors="replace",
                encoding="utf16",
            )
            return name
        except exceptions.InvalidAddressException as e:
            vollog.debug(
                f"Failed to get resident file content due to {e.__class__.__name__}"
            )
            return None

    def get_resident_filecontent(self) -> Optional[objects.Bytes]:
        """
        Returns the file content that is resident within this MFT attribute,
        for either the primary or an alternate data stream.
        """
        # smear observed in mass testing of samples
        # 4MB chosen as cutoff instead of 4KB to allow for recovery from format /L created file systems
        if (
            self.Attr_Header.ContentOffset > 0x400000
            or self.Attr_Header.ContentLength > 0x400000
        ):
            return None

        # To get the resident content, we jump to relative content offset and read name length * 2 bytes of data
        try:
            bytesobj = self._context.object(
                self.vol.type_name.split(constants.BANG)[0] + constants.BANG + "bytes",
                layer_name=self.vol.layer_name,
                offset=self.vol.offset + self.Attr_Header.ContentOffset,
                native_layer_name=self.vol.native_layer_name,
                length=self.Attr_Header.ContentLength,
            )
            return bytesobj
        except exceptions.InvalidAddressException as e:
            vollog.debug(
                f"Failed to get resident file content due to {e.__class__.__name__}"
            )
            return None



# This file is Copyright 2020 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
import ntpath
import re
from typing import Generator, List, Optional, Tuple, Type

from volatility3.framework import constants, exceptions, interfaces, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.renderers import format_hints
from volatility3.plugins.windows import handles, pslist

vollog = logging.getLogger(__name__)

FILE_DEVICE_DISK = 0x7
FILE_DEVICE_NETWORK_FILE_SYSTEM = 0x14
EXTENSION_CACHE_MAP = {
    "dat": "DataSectionObject",
    "img": "ImageSectionObject",
    "vacb": "SharedCacheMap",
}


class DumpFiles(interfaces.plugins.PluginInterface):
    """Dumps cached file contents from Windows memory samples."""

    _required_framework_version = (2, 0, 0)
    _version = (1, 0, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        # Since we're calling the plugin, make sure we have the plugin's requirements
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.IntRequirement(
                name="pid",
                description="Process ID to include (all other processes are excluded)",
                optional=True,
            ),
            requirements.ListRequirement(
                name="virtaddr",
                element_type=int,
                description="Dump the _FILE_OBJECTs at the given virtual address(es)",
                optional=True,
            ),
            requirements.ListRequirement(
                name="physaddr",
                element_type=int,
                description="Dump a single _FILE_OBJECTs at the given physical address(es)",
                optional=True,
            ),
            requirements.StringRequirement(
                name="filter",
                description="Dump files matching regular expression FILTER",
                optional=True,
            ),
            requirements.BooleanRequirement(
                name="ignore-case",
                description="Ignore case in filter match",
                default=False,
                optional=True,
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.VersionRequirement(
                name="handles", component=handles.Handles, version=(4, 0, 0)
            ),
        ]

    @classmethod
    def dump_file_producer(
        cls,
        file_object: interfaces.objects.ObjectInterface,
        memory_object: interfaces.objects.ObjectInterface,
        open_method: Type[interfaces.plugins.FileHandlerInterface],
        layer: interfaces.layers.DataLayerInterface,
        desired_file_name: str,
    ) -> Optional[interfaces.plugins.FileHandlerInterface]:
        """Produce a file from the memory object's get_available_pages() interface.

        :param file_object: the parent _FILE_OBJECT
        :param memory_object: the _CONTROL_AREA or _SHARED_CACHE_MAP
        :param open_method: class for constructing output files
        :param layer: the memory layer to read from
        :param desired_file_name: name of the output file
        :return: result status
        """
        filedata = open_method(desired_file_name)
        # Description of these variables:
        #   memoffset: offset in the specified layer where the page begins
        #   fileoffset: write to this offset in the destination file
        #   datasize: size of the page

        # track number of bytes written so we don't write empty files to disk
        bytes_written = 0
        try:
            for memoffset, fileoffset, datasize in memory_object.get_available_pages():
                data = layer.read(memoffset, datasize, pad=True)
                bytes_written += len(data)
                filedata.seek(fileoffset)
                filedata.write(data)
        except exceptions.InvalidAddressException:
            vollog.debug(f"Unable to dump file at {file_object.vol.offset:#x}")
            return None
        if not bytes_written:
            vollog.debug(
                f"No data is cached for the file at {file_object.vol.offset:#x}"
            )
            return None

        vollog.debug(f"Stored {filedata.preferred_filename}")
        return filedata

    @classmethod
    def process_file_object(
        cls,
        context: interfaces.context.ContextInterface,
        primary_layer_name: str,
        open_method: Type[interfaces.plugins.FileHandlerInterface],
        file_obj: interfaces.objects.ObjectInterface,
    ) -> Generator[Tuple, None, None]:
        """Given a FILE_OBJECT, dump data to separate files for each of the three file caches.

        :param context: the context to operate upon
        :param primary_layer_name: primary/virtual layer to operate on
        :param open_method: class for constructing output files
        :param file_obj: the FILE_OBJECT
        """
        # Filtering by these types of devices prevents us from processing other types of devices that
        # use the "File" object type, such as \Device\Tcp and \Device\NamedPipe.
        if file_obj.DeviceObject.DeviceType not in [
            FILE_DEVICE_DISK,
            FILE_DEVICE_NETWORK_FILE_SYSTEM,
        ]:
            vollog.log(
                constants.LOGLEVEL_VVV,
                f"The file object at {file_obj.vol.offset:#x} is not a file on disk",
            )
            return None

        # Depending on the type of object (DataSection, ImageSection, SharedCacheMap) we may need to
        # read from the memory layer or the primary layer.
        memory_layer_name = context.layers[primary_layer_name].config["memory_layer"]
        memory_layer = context.layers[memory_layer_name]
        primary_layer = context.layers[primary_layer_name]

        obj_name = file_obj.file_name_with_device()

        # This stores a list of tuples, describing what to dump and how to dump it.
        # Ex: (
        #     memory_object with get_available_pages() API (either CONTROL_AREA or SHARED_CACHE_MAP),
        #     layer to read from,
        #     file extension to apply,
        #     )
        dump_parameters = list()

        # The DataSectionObject and ImageSectionObject caches are handled in basically the same way.
        # We carve these "pages" from the memory_layer.
        for member_name, extension in [
            ("DataSectionObject", "dat"),
            ("ImageSectionObject", "img"),
        ]:
            try:
                section_obj = getattr(file_obj.SectionObjectPointer, member_name)
                control_area = section_obj.dereference().cast("_CONTROL_AREA")
                if control_area.is_valid():
                    dump_parameters.append((control_area, memory_layer, extension))
            except exceptions.InvalidAddressException:
                vollog.log(
                    constants.LOGLEVEL_VVV,
                    f"{member_name} is unavailable for file {file_obj.vol.offset:#x}",
                )

        # The SharedCacheMap is handled differently than the caches above.
        # We carve these "pages" from the primary_layer.
        try:
            scm_pointer = file_obj.SectionObjectPointer.SharedCacheMap
            shared_cache_map = scm_pointer.dereference().cast("_SHARED_CACHE_MAP")
            if shared_cache_map.is_valid():
                dump_parameters.append((shared_cache_map, primary_layer, "vacb"))
        except exceptions.InvalidAddressException:
            vollog.log(
                constants.LOGLEVEL_VVV,
                f"SharedCacheMap is unavailable for file {file_obj.vol.offset:#x}",
            )

        for memory_object, layer, extension in dump_parameters:
            cache_name = EXTENSION_CACHE_MAP[extension]
            desired_file_name = f"file.{file_obj.vol.offset:#x}.{memory_object.vol.offset:#x}.{cache_name}.{ntpath.basename(obj_name)}.{extension}"

            file_handle = cls.dump_file_producer(
                file_obj, memory_object, open_method, layer, desired_file_name
            )

            file_output = "Error dumping file"
            if file_handle:
                file_handle.close()
                file_output = file_handle.preferred_filename

            yield (
                cache_name,
                format_hints.Hex(file_obj.vol.offset),
                ntpath.basename(
                    obj_name
                ),  # temporary, so its easier to visualize output
                file_output,
            )

    def _generator(self, procs: List, offsets: List):
        kernel = self.context.modules[self.config["kernel"]]
        file_re = None
        if self.config["filter"]:
            flags = re.I if self.config["ignore-case"] else 0
            file_re = re.compile(self.config["filter"], flags)

        if procs:
            # The handles plugin doesn't expose any staticmethod/classmethod, and it also requires stashing
            # private variables, so we need an instance (for now, anyway). We _could_ call Handles._generator()
            # to do some of the other work that is duplicated here, but then we'd need to parse the TreeGrid
            # results instead of just dealing with them as direct objects here.
            type_map = handles.Handles.get_type_map(
                context=self.context,
                kernel_module_name=self.config["kernel"],
            )
            cookie = handles.Handles.find_cookie(
                context=self.context,
                kernel_module_name=self.config["kernel"],
            )

            dumped_files = set()

            for proc in procs:
                try:
                    object_table = proc.ObjectTable
                except exceptions.InvalidAddressException:
                    vollog.log(
                        constants.LOGLEVEL_VVV,
                        f"Cannot access _EPROCESS.ObjectTable at {proc.vol.offset:#x}",
                    )
                    continue

                for entry in handles.Handles.handles(
                    context=self.context,
                    kernel_module_name=self.config["kernel"],
                    handle_table=object_table,
                ):
                    try:
                        obj_type = entry.get_object_type(type_map, cookie)
                        if obj_type == "File":
                            file_obj = entry.Body.cast("_FILE_OBJECT")

                            if file_re:
                                name = file_obj.file_name_with_device()
                                if isinstance(name, renderers.UnreadableValue):
                                    continue
                                if not file_re.search(name):
                                    continue

                            if file_obj.vol.offset in dumped_files:
                                continue
                            dumped_files.add(file_obj.vol.offset)

                            for result in self.process_file_object(
                                self.context, kernel.layer_name, self.open, file_obj
                            ):
                                yield (0, result)
                    except exceptions.InvalidAddressException:
                        vollog.log(
                            constants.LOGLEVEL_VVV,
                            f"Cannot extract file from _OBJECT_HEADER at {entry.vol.offset:#x}",
                        )

                # Pull file objects from the VADs. This will produce DLLs and EXEs that are
                # mapped into the process as images, but that the process doesn't have an
                # explicit handle remaining open to those files on disk.
                for vad in proc.get_vad_root().traverse():
                    try:
                        if vad.has_member("ControlArea"):
                            # Windows xp and 2003
                            file_obj = vad.ControlArea.FilePointer.dereference()
                        elif vad.has_member("Subsection"):
                            # Vista and beyond
                            file_obj = vad.Subsection.ControlArea.FilePointer.dereference().cast(
                                "_FILE_OBJECT"
                            )
                        else:
                            continue

                        if not file_obj.is_valid():
                            continue

                        if file_re:
                            name = file_obj.file_name_with_device()
                            if isinstance(name, renderers.UnreadableValue):
                                continue
                            if not file_re.search(name):
                                continue

                        if file_obj.vol.offset in dumped_files:
                            continue
                        dumped_files.add(file_obj.vol.offset)

                        for result in self.process_file_object(
                            self.context, kernel.layer_name, self.open, file_obj
                        ):
                            yield (0, result)
                    except exceptions.InvalidAddressException:
                        vollog.log(
                            constants.LOGLEVEL_VVV,
                            f"Cannot extract file from VAD at {vad.vol.offset:#x}",
                        )

        elif offsets:
            virtual_layer_name = kernel.layer_name

            # FIXME - change this after standard access to physical layer
            physical_layer_name = self.context.layers[virtual_layer_name].config[
                "memory_layer"
            ]

            # Now process any offsets explicitly requested by the user.
            for offset, is_virtual in offsets:
                try:
                    file_obj = self.context.object(
                        kernel.symbol_table_name + constants.BANG + "_FILE_OBJECT",
                        layer_name=(
                            virtual_layer_name if is_virtual else physical_layer_name
                        ),
                        native_layer_name=virtual_layer_name,
                        offset=offset,
                    )
                    for result in self.process_file_object(
                        self.context, virtual_layer_name, self.open, file_obj
                    ):
                        yield (0, result)
                except exceptions.InvalidAddressException:
                    vollog.log(
                        constants.LOGLEVEL_VVV, f"Cannot extract file at {offset:#x}"
                    )

    def run(self):
        # a list of tuples (<int>, <bool>) where <int> is the address and <bool> is True for virtual.
        offsets = list()
        # a list of processes matching the pid filter. all files for these process(es) will be dumped.
        procs = list()

        if self.config["filter"] and (
            self.config["virtaddr"] or self.config["physaddr"]
        ):
            raise ValueError("Cannot use filter flag with an address flag")

        if self.config.get("virtaddr"):
            for virtaddr in self.config["virtaddr"]:
                offsets.append((virtaddr, True))

        if self.config.get("physaddr"):
            for physaddr in self.config["physaddr"]:
                offsets.append((physaddr, False))

        if not offsets:
            filter_func = pslist.PsList.create_pid_filter(
                [self.config.get("pid", None)]
            )
            procs = pslist.PsList.list_processes(
                context=self.context,
                kernel_module_name=self.config["kernel"],
                filter_func=filter_func,
            )

        return renderers.TreeGrid(
            [
                ("Cache", str),
                ("FileObject", format_hints.Hex),
                ("FileName", str),
                ("Result", str),
            ],
            self._generator(procs, offsets),
        )



import logging
import re
from typing import Any, List, Optional

from volatility3.framework import constants, interfaces

vollog = logging.getLogger(__name__)


class CLIFilter:
    def __init__(self, treegrid, filters: List[str]):
        self._filters = self._prepare(treegrid, filters)

    def _prepare(self, treegrid: interfaces.renderers.TreeGrid, filters: List[str]):
        """Runs through the filter strings and creates the necessary filter objects"""
        output = []

        for filter in filters:
            exclude = False
            regex = False
            pattern = None
            column_name = None
            if filter.startswith("-"):
                exclude = True
                filter = filter[1:]
            elif filter.startswith("+"):
                filter = filter[1:]
            components = filter.split(",")
            if len(components) < 2:
                pattern = components[0]
            else:
                column_name = components[0]
                pattern = ",".join(components[1:])
            if pattern and pattern.endswith("!"):
                regex = True
                pattern = pattern[:-1]
            column_num = None
            if column_name:
                for num, column in enumerate(treegrid.columns):
                    if column_name.lower() in column.name.lower():
                        column_num = num
                        break
            if pattern:
                output.append(ColumnFilter(column_num, pattern, regex, exclude))

        vollog.log(constants.LOGLEVEL_VVV, "Filters:\n" + repr(output))

        return output

    def filter(
        self,
        row: List[Any],
    ) -> bool:
        """Filters the row based on each of the column_filters"""
        if not self._filters:
            return False
        found = any(column_filter.found(row) for column_filter in self._filters)
        return not found


class ColumnFilter:
    def __init__(
        self,
        column_num: Optional[int],
        pattern: str,
        regex: bool = False,
        exclude: bool = False,
    ) -> None:
        self.column_num = column_num
        self.pattern = pattern
        self.regex = regex
        self.exclude = exclude

    def find(self, item) -> bool:
        """Identifies whether an item is found in the appropriate column"""
        try:
            if self.regex:
                return bool(re.search(self.pattern, f"{item}"))
            return self.pattern in f"{item}"
        except OSError:
            return False

    def found(self, row: List[Any]) -> bool:
        """Determines whether a row should be filtered

        If the classes exclude value is false, and the necessary pattern is found, the row is not filtered,
        otherwise it is filtered.
        """
        if self.column_num is None:
            found = any(self.find(x) for x in row)
        else:
            found = self.find(row[self.column_num])
        if self.exclude:
            return not found
        return found

    def __repr__(self) -> str:
        """Returns a display of a column filter"""
        return f"ColumnFilter(column={self.column_num},exclude={self.exclude},regex={self.regex},pattern={self.pattern})"



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging

from typing import List, Optional, Tuple

from volatility3.framework import interfaces, exceptions, symbols
from volatility3.framework.configuration import requirements
from volatility3.framework.symbols.windows import versions
from volatility3.plugins.windows import svcscan, pslist
from volatility3.framework.layers import scanners

vollog = logging.getLogger(__name__)


class SvcList(svcscan.SvcScan):
    """Lists services contained with the services.exe doubly linked list of services"""

    _required_framework_version = (2, 0, 0)

    # 2.0.0 - service_list signature changed
    _version = (2, 0, 0)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._enumeration_method = self.service_list

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        # Since we're calling the plugin, make sure we have the plugin's requirements
        return [
            requirements.VersionRequirement(
                name="svcscan", component=svcscan.SvcScan, version=(4, 0, 0)
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="bytes_scanner",
                component=scanners.BytesScanner,
                version=(1, 0, 0),
            ),
        ]

    @classmethod
    def _get_exe_range(cls, proc) -> Optional[Tuple[int, int]]:
        """
        Returns a tuple of starting address and size of
        the VAD containing services.exe
        """

        vad_root = proc.get_vad_root()
        for vad in vad_root.traverse():
            filename = vad.get_file_name()
            if isinstance(filename, str) and filename.lower().endswith(
                "\\services.exe"
            ):
                return [(vad.get_start(), vad.get_size())]

        return None

    @classmethod
    def service_list(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
        service_table_name: str,
        service_binary_dll_map,
        filter_func,
    ):
        kernel = context.modules[kernel_module_name]

        if not symbols.symbol_table_is_64bit(
            context=context, symbol_table_name=kernel.symbol_table_name
        ) or not versions.is_win10_15063_or_later(
            context=context, symbol_table=kernel.symbol_table_name
        ):
            vollog.warning(
                "This plugin only supports Windows 10 version 15063+ 64bit Windows memory samples"
            )
            return

        for proc in pslist.PsList.list_processes(
            context=context,
            kernel_module_name=kernel_module_name,
            filter_func=filter_func,
        ):
            try:
                proc_layer_name = proc.add_process_layer()
            except exceptions.InvalidAddressException:
                vollog.warning(
                    f"Unable to access memory of services.exe running with PID: {proc.UniqueProcessId}"
                )
                continue

            proc_layer = context.layers[proc_layer_name]

            exe_range = cls._get_exe_range(proc)
            if not exe_range:
                vollog.warning(
                    "Could not find the application executable VAD for services.exe. Unable to proceed."
                )
                continue

            for offset in proc_layer.scan(
                context=context,
                scanner=scanners.BytesScanner(needle=b"Sc27"),
                sections=exe_range,
            ):
                yield from cls.enumerate_vista_or_later_header(
                    context,
                    service_table_name,
                    service_binary_dll_map,
                    proc_layer_name,
                    offset,
                )



# This file is Copyright 2025 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

# Public researches: https://i.blackhat.com/USA21/Wednesday-Handouts/us-21-Fixing-A-Memory-Forensics-Blind-Spot-Linux-Kernel-Tracing-wp.pdf

import logging
from dataclasses import dataclass
from typing import Iterable, List, Optional

import volatility3.framework.symbols.linux.utilities.modules as linux_utilities_modules
from volatility3.framework import constants, exceptions, interfaces, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.constants import architectures
from volatility3.framework.objects import utility
from volatility3.framework.renderers import format_hints

vollog = logging.getLogger(__name__)


@dataclass
class ParsedTracepointFunc:
    """Parsed tracepoint_func struct, containing a selection of forensics valuable
    information."""

    tracepoint_name: str
    tracepoint_address: int
    probe_name: str
    probe_address: int
    probe_priority: int
    module_name: str
    module_address: int


class CheckTracepoints(interfaces.plugins.PluginInterface):
    """Detect tracepoints hooking

    Investigate the tracepoints subsystem to uncover kernel attached probes, which can be leveraged
    to hook kernel functions and modify their behaviour."""

    _version = (2, 0, 0)
    _required_framework_version = (2, 19, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=architectures.LINUX_ARCHS,
            ),
            requirements.VersionRequirement(
                name="linux_utilities_modules",
                component=linux_utilities_modules.Modules,
                version=(3, 0, 0),
            ),
            requirements.VersionRequirement(
                name="linux_utilities_module_gatherers",
                component=linux_utilities_modules.ModuleGatherers,
                version=(1, 0, 0),
            ),
        ]

    @classmethod
    def iterate_tracepoint_funcs(
        cls,
        context: interfaces.context.ContextInterface,
        layer_name: str,
        tracepoint: interfaces.objects.ObjectInterface,
    ) -> Optional[Iterable[interfaces.objects.ObjectInterface]]:
        """Extract probes represented by tracepoint_func structs from a
        tracepoint funcs member.

        Args:
            tracepoint: The tracepoint struct to parse

        Yields:
            An iterable of tracepoint_func structs
        """

        layer = context.layers[layer_name]
        # Ignore tracepoints without attached probes
        if not tracepoint.funcs.is_readable():
            return None

        current_tracepoint_func = tracepoint.funcs.dereference()
        # Inspired by kernel's debug_print_probes()
        while (
            layer.is_valid(current_tracepoint_func.vol.offset)
            and current_tracepoint_func.func.is_readable()
        ):
            yield current_tracepoint_func
            current_tracepoint_func = context.object(
                tracepoint.get_symbol_table_name() + constants.BANG + "tracepoint_func",
                layer_name,
                current_tracepoint_func.vol.offset + current_tracepoint_func.vol.size,
            )

    @classmethod
    def parse_tracepoint(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
        known_modules: List[linux_utilities_modules.ModuleInfo],
        tracepoint: interfaces.objects.ObjectInterface,
        run_hidden_modules: bool = True,
    ) -> Optional[Iterable[ParsedTracepointFunc]]:
        """Parse a tracepoint struct to highlight tracepoints kernel hooking.

        Args:
            known_modules: A dict of known modules, used to locate callbacks origin. Typically obtained through run_modules_scanners().
            tracepoint: The tracepoint struct to parse
            run_hidden_modules: Whether to run the hidden_modules plugin or not. Note: it won't be run, even if specified, \
            if the "hidden_modules" key is present in known_modules.

        Yields:
            An iterable of ParsedTracepointFunc dataclasses, containing a selection of useful fields related to a tracepoint struct
        """
        kernel = context.modules[kernel_module_name]

        for tracepoint_func in cls.iterate_tracepoint_funcs(
            context, kernel.layer_name, tracepoint
        ):
            try:
                tracepoint_name = utility.pointer_to_string(tracepoint.name, count=512)
            except exceptions.InvalidAddressException:
                vollog.debug(
                    f"Tracepoint function at {tracepoint.vol.offset:#x} is smeared."
                )
                continue

            probe_handler_address = tracepoint_func.func
            probe_handler_symbol = module_address = module_name = None

            # Try to lookup within the known modules if the probe_handler address fits
            mod_info, probe_handler_symbol = (
                linux_utilities_modules.Modules.module_lookup_by_address(
                    context,
                    kernel_module_name,
                    known_modules,
                    probe_handler_address,
                )
            )

            # Fetch more information about the module
            if mod_info is not None:
                module_address = mod_info.offset
                module_name = mod_info.name
            else:
                vollog.debug(
                    f"Could not determine tracepoint@{tracepoint.vol.offset:#x} probe handler {probe_handler_address:#x} module origin.",
                )

            if hasattr(tracepoint_func, "prio"):
                prio = tracepoint_func.prio
            else:
                prio = None

            yield ParsedTracepointFunc(
                tracepoint_name,
                tracepoint.vol.offset,
                probe_handler_symbol,
                probe_handler_address,
                prio,
                module_name,
                module_address,
            )

    @classmethod
    def iterate_tracepoints_array(
        cls, context: interfaces.context.ContextInterface, kernel_name: str
    ) -> List[interfaces.objects.ObjectInterface]:
        """Iterate over (tracepoint_ptr_t *)__start___tracepoints_ptrs.
        Handles CONFIG_HAVE_ARCH_PREL32_RELOCATIONS.

        Returns:
            A list of tracepoint structs
        """

        kernel = context.modules[kernel_name]

        tracepoints = []
        tracepoints_start = kernel.object_from_symbol("__start___tracepoints_ptrs")
        tracepoints_end = kernel.get_absolute_symbol_address(
            "__stop___tracepoints_ptrs"
        )
        tracepoints_array_size = tracepoints_end - tracepoints_start.vol.offset
        # kernel's tracepoint_ptr_deref() and tracepoint_ptr_t
        # adjust depending on the use of PC-relative addressing
        # or not.
        # Relocation is commonly used to store pointers as offsets
        # relative to their own address rather than absolute addresses/pointers.
        config_have_arch_prel32_relocations = (
            tracepoints_start.vol.subtype.type_name
            == kernel.symbol_table_name + constants.BANG + "int"
        )
        if config_have_arch_prel32_relocations:
            tracepoints_relative_offsets = tracepoints_start.cast(
                "array",
                count=tracepoints_array_size // kernel.get_type("int").size,
                subtype=kernel.get_type("int"),
            )
            for relative_offset in tracepoints_relative_offsets:
                # relative_offset is the value stored at relative_offset.vol.offset
                # See kernel's offset_to_ptr(). Example:
                # 0xffff9da125e0 = 0x7af138        + 0xffff9d2634a8
                absolute_address = relative_offset + relative_offset.vol.offset
                tracepoint = kernel.object(
                    "tracepoint",
                    absolute_address,
                    absolute=True,
                )
                tracepoints.append(tracepoint)
        else:
            tracepoints = utility.array_of_pointers(
                tracepoints_start,
                tracepoints_array_size // kernel.get_type("pointer").size,
                kernel.symbol_table_name + constants.BANG + "tracepoint",
                context,
            )

        return tracepoints

    def _generator(self):
        kernel_name = self.config["kernel"]
        kernel = self.context.modules[kernel_name]
        kernel_layer = self.context.layers[kernel.layer_name]

        if not kernel.has_symbol("__start___tracepoints_ptrs"):
            vollog.error(
                'The provided symbol table does not include the "__start___tracepoints_ptrs" symbol.'
                "This means you are either analyzing an unsupported kernel version or that your symbol table is corrupted."
            )
            return

        known_modules = linux_utilities_modules.Modules.run_modules_scanners(
            context=self.context,
            kernel_module_name=self.config["kernel"],
            caller_wanted_gatherers=linux_utilities_modules.ModuleGatherers.all_gatherers_identifier,
        )
        tracepoints = self.iterate_tracepoints_array(self.context, kernel_name)

        for tracepoint in tracepoints:
            if not kernel_layer.is_valid(tracepoint.vol.offset):
                continue

            for tracepoint_parsed in self.parse_tracepoint(
                self.context, kernel_name, known_modules, tracepoint
            ):
                formatted_results = (
                    tracepoint_parsed.tracepoint_name,
                    format_hints.Hex(tracepoint_parsed.tracepoint_address),
                    tracepoint_parsed.probe_name or renderers.NotAvailableValue(),
                    format_hints.Hex(tracepoint_parsed.probe_address),
                    tracepoint_parsed.probe_priority or renderers.NotAvailableValue(),
                    tracepoint_parsed.module_name or renderers.NotAvailableValue(),
                    (
                        format_hints.Hex(tracepoint_parsed.module_address)
                        if tracepoint_parsed.module_address is not None
                        else renderers.NotAvailableValue()
                    ),
                )
                yield (
                    0,
                    formatted_results,
                )

    def run(self):
        columns = [
            ("tracepoint", str),
            ("tracepoint address", format_hints.Hex),
            ("Probe", str),
            ("Probe address", format_hints.Hex),
            ("Probe priority", int),
            ("Module", str),
            ("Module address", format_hints.Hex),
        ]

        return renderers.TreeGrid(
            columns,
            self._generator(),
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
from typing import Generator, Iterable, List, Optional, Dict, Tuple

from volatility3.framework import symbols, constants, exceptions, interfaces, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.renderers import format_hints
from volatility3.framework.symbols import intermed
from volatility3.framework.symbols.windows.extensions import pe
from volatility3.plugins.windows import pedump, pslist

vollog = logging.getLogger(__name__)


class Modules(interfaces.plugins.PluginInterface):
    """Lists the loaded kernel modules."""

    _required_framework_version = (2, 0, 0)

    # 3.0.0 - changed signature of get_session_layers, added get_session_layers_map
    _version = (3, 0, 0)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._enumeration_method = self.list_modules

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.VersionRequirement(
                name="pedump", component=pedump.PEDump, version=(2, 0, 0)
            ),
            requirements.BooleanRequirement(
                name="dump",
                description="Extract listed modules",
                default=False,
                optional=True,
            ),
            requirements.IntRequirement(
                name="base",
                description="Extract a single module with BASE address",
                optional=True,
            ),
            requirements.StringRequirement(
                name="name",
                description="module name/sub string",
                optional=True,
                default=None,
            ),
        ]

    def dump_module(self, session_layers, pe_table_name, mod):
        session_layer_name = self.find_session_layer(
            self.context, session_layers, mod.DllBase
        )
        file_output = f"Cannot find a viable session layer for {mod.DllBase:#x}"
        if session_layer_name:
            file_output = pedump.PEDump.dump_ldr_entry(
                self.context,
                pe_table_name,
                mod,
                self.open,
                layer_name=session_layer_name,
            )
            if not file_output:
                file_output = "Error outputting file"

        return file_output

    def _generator(self):
        pe_table_name = None
        session_layers = None

        if self.config["dump"]:
            pe_table_name = intermed.IntermediateSymbolTable.create(
                self.context,
                self.config_path,
                "windows",
                "pe",
                class_types=pe.class_types,
            )

            session_layers = list(
                self.get_session_layers(
                    context=self.context,
                    kernel_module_name=self.config["kernel"],
                )
            )

        for mod in self._enumeration_method(
            self.context, kernel_module_name=self.config["kernel"]
        ):
            if self.config["base"] and self.config["base"] != mod.DllBase:
                continue

            try:
                BaseDllName = mod.BaseDllName.get_string()
                if self.config["name"] and self.config["name"] not in BaseDllName:
                    continue
            except exceptions.InvalidAddressException:
                BaseDllName = interfaces.renderers.BaseAbsentValue()

            try:
                FullDllName = mod.FullDllName.get_string()
            except exceptions.InvalidAddressException:
                FullDllName = interfaces.renderers.BaseAbsentValue()

            file_output = "Disabled"
            if self.config["dump"]:
                file_output = self.dump_module(session_layers, pe_table_name, mod)

            yield (
                0,
                (
                    format_hints.Hex(mod.vol.offset),
                    format_hints.Hex(mod.DllBase),
                    format_hints.Hex(mod.SizeOfImage),
                    BaseDllName,
                    FullDllName,
                    file_output,
                ),
            )

    @classmethod
    def get_kernel_space_start(cls, context, module_name: str) -> int:
        """
        Returns the starting address of the kernel address space

        This method allows plugins that analyze kernel data structures to quickly detect
        smeared or otherwise invalid data as many pointers must point into the kernel or
        access during runtime would crash the system
        """
        module = context.modules[module_name]

        # default is used if/when MmSystemRangeStart is paged out
        if symbols.symbol_table_is_64bit(
            context=context, symbol_table_name=module.symbol_table_name
        ):
            object_type = "unsigned long long"
            default_start = 0xFFFF800000000000
        else:
            object_type = "unsigned long"
            default_start = 0x80000000

        range_start_offset = module.get_symbol("MmSystemRangeStart").address

        try:
            kernel_space_start = module.object(
                object_type=object_type, offset=range_start_offset
            )
        except exceptions.InvalidAddressException:
            vollog.debug(
                f"Unable to read MmSystemRangeStart. Defaulting to {default_start:#x} for the kernel space start."
            )
            kernel_space_start = default_start

        layer = context.layers[module.layer_name]

        return kernel_space_start & layer.address_mask

    @classmethod
    def _do_get_session_layers(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
        pids: Optional[List[int]] = None,
    ) -> Generator[Tuple[int, str], None, None]:
        """Build a cache of possible virtual layers, in priority starting with
        the primary/kernel layer. Then keep one layer per session by cycling
        through the process list.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            kernel_module_name: The name of the module for the kernel
            pids: A list of process identifiers to include exclusively or None for no filter

        Returns:
            A generator of session layer names
        """
        seen_ids: List[interfaces.objects.ObjectInterface] = []
        filter_func = pslist.PsList.create_pid_filter(pids or [])

        kernel = context.modules[kernel_module_name]

        for proc in pslist.PsList.list_processes(
            context=context,
            kernel_module_name=kernel_module_name,
            filter_func=filter_func,
        ):
            proc_id = "Unknown"
            try:
                proc_id = proc.UniqueProcessId
                proc_layer_name = proc.add_process_layer()

                # create the session space object in the process' own layer.
                # not all processes have a valid session pointer.
                try:
                    session_space = context.object(
                        kernel.symbol_table_name + constants.BANG + "_MM_SESSION_SPACE",
                        layer_name=kernel.layer_name,
                        offset=proc.Session,
                    )
                    session_id = session_space.SessionId

                except exceptions.SymbolError:
                    # In Windows 11 24H2, the _MM_SESSION_SPACE type was
                    # replaced with _PSP_SESSION_SPACE, and the kernel PDB
                    # doesn't contain information about its members (otherwise,
                    # we would just fall back to the new type). However, it
                    # appears to be, for our purposes, functionally identical
                    # to the _MM_SESSION_SPACE. Because _MM_SESSION_SPACE
                    # stores its session ID at offset 8 as an unsigned long, we
                    # create an unsigned long at that offset and use that
                    # instead.
                    session_id = context.object(
                        layer_name=kernel.layer_name,
                        object_type=kernel.symbol_table_name
                        + constants.BANG
                        + "unsigned long",
                        offset=proc.Session + 8,
                    )

                if session_id in seen_ids:
                    continue

            except exceptions.InvalidAddressException:
                vollog.log(
                    constants.LOGLEVEL_VVV,
                    f"Process {proc_id} does not have a valid Session or a layer could not be constructed for it",
                )
                continue

            # save the layer if we haven't seen the session yet
            seen_ids.append(session_id)
            yield session_id, proc_layer_name

    @classmethod
    def get_session_layers(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
        pids: Optional[List[int]] = None,
    ) -> Generator[str, None, None]:
        """
        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            kernel_module_name: The name of the module for the kernel
            pids: A list of process identifiers to include exclusively or None for no filter

        Yields the names of the unique memory layers that map sessions
        """
        for _session_id, proc_layer_name in cls._do_get_session_layers(
            context, kernel_module_name, pids
        ):
            yield proc_layer_name

    @classmethod
    def get_session_layers_map(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
        pids: Optional[List[int]] = None,
    ) -> Dict[int, str]:
        """
        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            kernel_module_name: The name of the module for the kernel
            pids: A list of process identifiers to include exclusively or None for no filter

        Wraps `_do_get_session_layers` to produce a dictionary where each key is a session_id
        and the value is the name of the layer for that session
        """
        return dict(cls._do_get_session_layers(context, kernel_module_name, pids))

    @classmethod
    def find_session_layer(
        cls,
        context: interfaces.context.ContextInterface,
        session_layers: Iterable[str],
        base_address: int,
    ):
        """Given a base address and a list of layer names, find a layer that
        can access the specified address.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            layer_name: The name of the layer on which to operate
            symbol_table: The name of the table containing the kernel symbols
            session_layers: A list of session layer names
            base_address: The base address to identify the layers that can access it

        Returns:
            Layer name or None if no layers that contain the base address can be found
        """

        for layer_name in session_layers:
            if context.layers[layer_name].is_valid(base_address):
                return layer_name

        return None

    @classmethod
    def list_modules(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
    ) -> Iterable[interfaces.objects.ObjectInterface]:
        """Lists all the modules in the primary layer.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            kernel_module_name: The name of the module for the kernel
        Returns:
            A list of Modules as retrieved from PsLoadedModuleList
        """

        kernel = context.modules[kernel_module_name]
        if not kernel.offset:
            raise ValueError(
                "Intel layer does not have an associated kernel virtual offset, failing"
            )

        try:
            # use this type if its available (starting with windows 10)
            ldr_entry_type = kernel.get_type("_KLDR_DATA_TABLE_ENTRY")
        except exceptions.SymbolError:
            ldr_entry_type = kernel.get_type("_LDR_DATA_TABLE_ENTRY")

        type_name = ldr_entry_type.type_name.split(constants.BANG)[1]

        list_head = kernel.get_symbol("PsLoadedModuleList").address
        list_entry = kernel.object(object_type="_LIST_ENTRY", offset=list_head)
        reloff = ldr_entry_type.relative_child_offset("InLoadOrderLinks")
        module = kernel.object(
            object_type=type_name, offset=list_entry.vol.offset - reloff, absolute=True
        )

        yield from module.InLoadOrderLinks

    def run(self):
        return renderers.TreeGrid(
            [
                ("Offset", format_hints.Hex),
                ("Base", format_hints.Hex),
                ("Size", format_hints.Hex),
                ("Name", str),
                ("Path", str),
                ("File output", str),
            ],
            self._generator(),
        )



# This file is Copyright 2022 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import functools
from typing import List, Optional, Tuple, Iterable

from volatility3.framework import exceptions, interfaces


class LinearlyMappedLayer(interfaces.layers.TranslationLayerInterface):
    """Class to differentiate Linearly Mapped layers (where a => b implies that
    a + c => b + c)"""

    ### Translation layer convenience function

    def translate(
        self, offset: int, ignore_errors: bool = False
    ) -> Tuple[Optional[int], Optional[str]]:
        mapping = list(self.mapping(offset, 0, ignore_errors))
        if len(mapping) == 1:
            original_offset, _, mapped_offset, _, layer = mapping[0]
            if original_offset != offset:
                raise exceptions.LayerException(
                    self.name, f"Layer {self.name} claims to map linearly but does not"
                )
        else:
            if ignore_errors:
                # We should only hit this if we ignored errors, but check anyway
                return None, None
            raise exceptions.InvalidAddressException(
                self.name, offset, f"Cannot translate {offset} in layer {self.name}"
            )
        return mapped_offset, layer

    # ## Read/Write functions for mapped pages
    # Redefine read here for speed reasons (so we don't call a processing method

    @functools.lru_cache(maxsize=512)
    def read(self, offset: int, length: int, pad: bool = False) -> bytes:
        """Reads an offset for length bytes and returns 'bytes' (not 'str') of
        length size."""
        current_offset = offset
        output: List[bytes] = []
        for offset, _, mapped_offset, mapped_length, layer in self.mapping(
            offset, length, ignore_errors=pad
        ):
            if not pad and offset > current_offset:
                raise exceptions.InvalidAddressException(
                    self.name,
                    current_offset,
                    f"Layer {self.name} cannot map offset: {current_offset}",
                )
            elif offset > current_offset:
                output += [b"\x00" * (offset - current_offset)]
                current_offset = offset
            elif offset < current_offset:
                raise exceptions.LayerException(
                    self.name, "Mapping returned an overlapping element"
                )
            if mapped_length > 0:
                output += [
                    self._context.layers.read(layer, mapped_offset, mapped_length, pad)
                ]
            current_offset += mapped_length
        recovered_data = b"".join(output)
        return recovered_data + b"\x00" * (length - len(recovered_data))

    def write(self, offset: int, value: bytes) -> None:
        """Writes a value at offset, distributing the writing across any
        underlying mapping."""
        current_offset = offset
        length = len(value)
        for offset, _, mapped_offset, length, layer in self.mapping(offset, length):
            if offset > current_offset:
                raise exceptions.InvalidAddressException(
                    self.name,
                    current_offset,
                    f"Layer {self.name} cannot map offset: {current_offset}",
                )
            elif offset < current_offset:
                raise exceptions.LayerException(
                    self.name, "Mapping returned an overlapping element"
                )
            self._context.layers.write(layer, mapped_offset, value[:length])
            value = value[length:]
            current_offset += length

    def _scan_iterator(
        self,
        scanner: "interfaces.layers.ScannerInterface",
        sections: Iterable[Tuple[int, int]],
        linear: bool = True,
    ) -> Iterable[interfaces.layers.IteratorValue]:
        return super()._scan_iterator(scanner, sections, linear)



import contextlib
import functools
import logging
import struct
from typing import Dict, List, Optional, Tuple, Union

from volatility3.plugins.windows import poolscanner

from volatility3.framework import (
    constants,
    exceptions,
    interfaces,
    objects,
    renderers,
    symbols,
)
from volatility3.framework.renderers import conversion

vollog = logging.getLogger(__name__)


class POOL_HEADER(objects.StructType):
    """A kernel pool allocation header.

    Exists at the base of the allocation and provides a tag that we can
    scan for.
    """

    def get_object(
        self,
        constraint: poolscanner.PoolConstraint,
        use_top_down: bool,
        kernel_symbol_table: Optional[str] = None,
        native_layer_name: Optional[str] = None,
    ) -> Optional[interfaces.objects.ObjectInterface]:
        """Carve an object or data structure from a kernel pool allocation

        Args:
            constraint: a PoolConstraint object used to get the pool allocation header object
            use_top_down: for delineating how a windows version finds the size of the object body
            kernel_symbol_table: in case objects of a different symbol table are scanned for
            native_layer_name: the name of the layer where the data originally lived

        Returns:
            An object as found from a POOL_HEADER
        """

        type_name = constraint.type_name
        executive = constraint.object_type is not None

        symbol_table_name = self.vol.type_name.split(constants.BANG)[0]
        if constants.BANG in type_name:
            symbol_table_name, type_name = type_name.split(constants.BANG)[0:2]

        # when checking for symbols from a table other than nt_symbols grab _OBJECT_HEADER from the kernel
        # because symbol_table_name will be different from kernel_symbol_table.
        if kernel_symbol_table:
            object_header_type = self._context.symbol_space.get_type(
                kernel_symbol_table + constants.BANG + "_OBJECT_HEADER"
            )
        else:
            # otherwise symbol_table_name *is* the kernel symbol table, so just use that.
            object_header_type = self._context.symbol_space.get_type(
                symbol_table_name + constants.BANG + "_OBJECT_HEADER"
            )

        pool_header_size = self.vol.size

        # if there is no object type, then just instantiate a structure
        if not executive:
            mem_object = self._context.object(
                symbol_table_name + constants.BANG + type_name,
                layer_name=self.vol.layer_name,
                offset=self.vol.offset + pool_header_size,
                native_layer_name=native_layer_name,
            )
            yield mem_object

        # otherwise we have an executive object in the pool
        else:
            if symbols.symbol_table_is_64bit(
                context=self._context, symbol_table_name=symbol_table_name
            ):
                alignment = 16
            else:
                alignment = 8

            # use the top down approach for windows 8 and later
            if use_top_down:
                body_offset = object_header_type.relative_child_offset("Body")
                infomask_offset = object_header_type.relative_child_offset("InfoMask")
                pointercount_offset = object_header_type.relative_child_offset(
                    "PointerCount"
                )
                pointercount_size = object_header_type.members["PointerCount"][1].size
                (
                    optional_headers,
                    lengths_of_optional_headers,
                ) = self._calculate_optional_header_lengths(
                    self._context, kernel_symbol_table
                )
                padding_available = (
                    None
                    if "PADDING_INFO" not in optional_headers
                    else optional_headers.index("PADDING_INFO")
                )
                max_optional_headers_length = sum(lengths_of_optional_headers)

                # define the starting and ending bounds for the scan
                start_offset = self.vol.offset + pool_header_size
                addr_limit = min(
                    max_optional_headers_length, self.BlockSize * alignment
                )

                # A single read is better than lots of little one-byte reads.
                # We're ok padding this, because the byte we'd check would be 0 which would only be valid if there
                # were no optional headers in the first place (ie, if we read too much for headers that don't exist,
                # but the bit we could read were valid)
                infomask_data = self._context.layers[self.vol.layer_name].read(
                    start_offset, addr_limit + infomask_offset, pad=True
                )

                # Addr stores the offset to the potential start of the OBJECT_HEADER from just after the POOL_HEADER
                # It will always be aligned to a particular alignment
                for addr in range(0, addr_limit, alignment):
                    infomask_value = infomask_data[addr + infomask_offset]
                    pointercount_value = int.from_bytes(
                        infomask_data[
                            addr + pointercount_offset : addr
                            + pointercount_offset
                            + pointercount_size
                        ],
                        byteorder="little",
                        signed=True,
                    )
                    if not 0x1000000 > pointercount_value >= 0:
                        continue

                    padding_present = False
                    optional_headers_length = 0
                    for i in range(len(lengths_of_optional_headers)):
                        if infomask_value & (1 << i):
                            optional_headers_length += lengths_of_optional_headers[i]
                            if i == padding_available:
                                padding_present = True

                    # PADDING_INFO is a special case (4 bytes that contain the total padding length)
                    padding_length = 0
                    if padding_present:
                        # Read the four bytes from just before the next optional_headers_length minus the padding_info size
                        #
                        #  ---------------
                        #  POOL_HEADER
                        #  ---------------
                        #
                        #  start of PADDING_INFO
                        #  ---------------
                        #  End of other optional headers
                        #  ---------------
                        #  OBJECT_HEADER
                        #  ---------------
                        if addr - optional_headers_length < 0:
                            continue
                        (padding_length,) = struct.unpack(
                            "<I",
                            infomask_data[
                                addr - optional_headers_length : addr
                                - optional_headers_length
                                + 4
                            ],
                        )
                        padding_length -= lengths_of_optional_headers[
                            padding_available or 0
                        ]

                    # Certain versions of windows have PADDING_INFO lengths that are too long
                    # So we now check that the padding length is at a minimum the right length
                    # and that it doesn't go beyond the entirety of the data
                    if addr - optional_headers_length >= padding_length > addr:
                        continue

                    with contextlib.suppress(
                        TypeError, exceptions.InvalidAddressException
                    ):
                        mem_object = self._context.object(
                            symbol_table_name + constants.BANG + type_name,
                            layer_name=self.vol.layer_name,
                            offset=addr + body_offset + start_offset,
                            native_layer_name=native_layer_name,
                        )

                        if mem_object.is_valid():
                            yield mem_object

            # use the bottom up approach for windows 7 and earlier
            else:
                type_size = self._context.symbol_space.get_type(
                    symbol_table_name + constants.BANG + type_name
                ).size
                if constraint.additional_structures:
                    for additional_structure in constraint.additional_structures:
                        type_size += self._context.symbol_space.get_type(
                            symbol_table_name + constants.BANG + additional_structure
                        ).size

                rounded_size = conversion.round(type_size, alignment, up=True)

                mem_object = self._context.object(
                    symbol_table_name + constants.BANG + type_name,
                    layer_name=self.vol.layer_name,
                    offset=self.vol.offset + self.BlockSize * alignment - rounded_size,
                    native_layer_name=native_layer_name,
                )

                with contextlib.suppress(TypeError, exceptions.InvalidAddressException):
                    if mem_object.is_valid():
                        yield mem_object

    @classmethod
    @functools.lru_cache
    def _calculate_optional_header_lengths(
        cls, context: interfaces.context.ContextInterface, symbol_table_name: str
    ) -> Tuple[List[str], List[int]]:
        headers = []
        sizes = []
        for header in [
            "CREATOR_INFO",
            "NAME_INFO",
            "HANDLE_INFO",
            "QUOTA_INFO",
            "PROCESS_INFO",
            "AUDIT_INFO",
            "EXTENDED_INFO",
            "HANDLE_REVOCATION_INFO",
            "PADDING_INFO",
        ]:
            with contextlib.suppress(AttributeError, exceptions.SymbolError):
                type_name = (
                    f"{symbol_table_name}{constants.BANG}_OBJECT_HEADER_{header}"
                )
                header_type = context.symbol_space.get_type(type_name)
                headers.append(header)
                sizes.append(header_type.size)
                # Some of these may not exist, for example:
                #   if build < 9200: PADDING_INFO else: AUDIT_INFO
                #   if build == 10586: HANDLE_REVOCATION_INFO else EXTENDED_INFO
                # based on what's present and what's not, this list should be the right order and the right length
        return headers, sizes

    def is_free_pool(self):
        return self.PoolType == 0

    def is_paged_pool(self):
        return self.PoolType % 2 == 0 and self.PoolType > 0

    def is_nonpaged_pool(self):
        return self.PoolType % 2 == 1


class POOL_HEADER_VISTA(POOL_HEADER):
    """A kernel pool allocation header, updated for Vista and later.

    Exists at the base of the allocation and provides a tag that we can
    scan for.
    """

    def is_paged_pool(self):
        return self.PoolType % 2 == 1

    def is_nonpaged_pool(self):
        return self.PoolType % 2 == 0 and self.PoolType > 0


class POOL_TRACKER_BIG_PAGES(objects.StructType):
    """A kernel big page pool tracker."""

    pool_type_lookup: Dict[str, str] = {}

    def _generate_pool_type_lookup(self):
        # Enumeration._generate_inverse_choices() raises ValueError because multiple enum names map to the same
        # value in the kernel _POOL_TYPE so create a custom mapping here and take the first match
        symbol_table_name = self.vol.type_name.split(constants.BANG)[0]
        pool_type_enum = self._context.symbol_space.get_enumeration(
            symbol_table_name + constants.BANG + "_POOL_TYPE"
        )
        for k, v in pool_type_enum.choices.items():
            if v not in self.pool_type_lookup:
                self.pool_type_lookup[v] = k

    def is_valid(self) -> bool:
        return self.Key > 0

    def is_free(self) -> bool:
        """Returns if the allocation is freed (True) or in-use (False)"""
        return self.Va & 1 == 1

    def get_key(self) -> str:
        """Returns the Key value as a 4 character string"""
        tag_bytes = objects.convert_value_to_data(
            self.Key, int, objects.DataFormatInfo(4, "little", False)
        )
        return "".join([chr(x) if 32 < x < 127 else "" for x in tag_bytes])

    def get_pool_type(self) -> Union[str, interfaces.renderers.BaseAbsentValue]:
        """Returns the enum name for the PoolType value on applicable systems"""
        # Not applicable until Vista
        if hasattr(self, "PoolType"):
            if not self.pool_type_lookup:
                self._generate_pool_type_lookup()
            return self.pool_type_lookup.get(
                self.PoolType, f"Unknown choice {self.PoolType}"
            )
        else:
            return renderers.NotApplicableValue()

    def get_number_of_bytes(self) -> Union[int, interfaces.renderers.BaseAbsentValue]:
        """Returns the NumberOfBytes value on applicable systems"""
        # Not applicable until Vista
        try:
            return self.NumberOfBytes
        except AttributeError:
            return renderers.NotApplicableValue()


class ExecutiveObject(interfaces.objects.ObjectInterface):
    """This is used as a "mixin" that provides all kernel executive objects
    with a means of finding their own object header."""

    def get_object_header(
        self, symbol_table_name: Optional[str] = None
    ) -> "OBJECT_HEADER":
        if constants.BANG not in self.vol.type_name:
            raise ValueError(
                f"Invalid symbol table name syntax (no {constants.BANG} found)"
            )

        # caller provided symbol table allows for scanning for objects from any module
        if not symbol_table_name:
            symbol_table_name = self.vol.type_name.split(constants.BANG)[0]

        body_offset = self._context.symbol_space.get_type(
            symbol_table_name + constants.BANG + "_OBJECT_HEADER"
        ).relative_child_offset("Body")
        return self._context.object(
            symbol_table_name + constants.BANG + "_OBJECT_HEADER",
            layer_name=self.vol.layer_name,
            offset=self.vol.offset - body_offset,
            native_layer_name=self.vol.native_layer_name,
        )

    def get_name(self, symbol_table_name: Optional[str] = None) -> Optional[str]:
        try:
            return self.get_object_header(symbol_table_name).get_name()
        except exceptions.InvalidAddressException:
            return None


class OBJECT_HEADER(objects.StructType):
    """A class for the headers for executive kernel objects, which contains
    quota information, ownership details, naming data, and ACLs."""

    def is_valid(self) -> bool:
        """Determine if the object is valid."""

        # if self.InfoMask > 0x48:
        #    return False

        try:
            if self.PointerCount > 0x1000000 or self.PointerCount < 0:
                return False
        except exceptions.InvalidAddressException:
            return False

        return True

    def get_object_type(
        self, type_map: Dict[int, str], cookie: Optional[int] = None
    ) -> Optional[str]:
        """Across all Windows versions, the _OBJECT_HEADER embeds details on
        the type of object (i.e. process, file) but the way its embedded
        differs between versions.

        This API abstracts away those details.
        """

        if self.vol.get("object_header_object_type", None) is not None:
            return self.vol.object_header_object_type

        try:
            # vista and earlier have a Type member
            length = self.Type.member("Name").Length
            if length == 0 or length > 128:
                string = None
            else:
                string = self.Type.Name.String
                if len(string) == 0 or len(string) > 128:
                    string = None

            self._vol["object_header_object_type"] = string

        except AttributeError:
            # windows 7 and later have a TypeIndex, but windows 10
            # further encodes the index value with nt1!ObHeaderCookie
            try:
                type_index = ((self.vol.offset >> 8) ^ cookie ^ self.TypeIndex) & 0xFF
            except (AttributeError, TypeError):
                type_index = self.TypeIndex

            self._vol["object_header_object_type"] = type_map.get(type_index)
        return self.vol.object_header_object_type

    @property
    def NameInfo(self) -> interfaces.objects.ObjectInterface:
        if constants.BANG not in self.vol.type_name:
            raise ValueError(
                f"Invalid symbol table name syntax (no {constants.BANG} found)"
            )

        symbol_table_name = self.vol.type_name.split(constants.BANG)[0]

        if symbol_table_name in self._context.modules:
            ntkrnlmp = self._context.modules[symbol_table_name]
        else:
            layer = self._context.layers[self.vol.native_layer_name]
            kvo = layer.config.get("kernel_virtual_offset", None)

            if kvo is None:
                raise AttributeError(
                    f"Could not find kernel_virtual_offset for layer: {self.vol.layer_name}"
                )

            # We know this symbol table name can't exist because we checked for it earlier
            ntkrnlmp = self._context.module(
                symbol_table_name, layer_name=self.vol.layer_name, offset=kvo
            )

        try:
            header_offset = self.NameInfoOffset
        except AttributeError:
            # http://codemachine.com/article_objectheader.html (Windows 7 and later)
            name_info_bit = 0x2

            address = ntkrnlmp.get_symbol("ObpInfoMaskToOffset").address
            calculated_index = self.InfoMask & (name_info_bit | (name_info_bit - 1))

            header_offset = ntkrnlmp.object(
                "unsigned char",
                layer_name=self.vol.native_layer_name,
                offset=address + calculated_index,
            )

        if header_offset == 0:
            raise ValueError(
                f"Could not find _OBJECT_HEADER_NAME_INFO for object at {self.vol.offset} of layer {self.vol.layer_name}"
            )

        header = ntkrnlmp.object(
            "_OBJECT_HEADER_NAME_INFO",
            layer_name=self.vol.layer_name,
            offset=self.vol.offset - header_offset,
            native_layer_name=self.vol.native_layer_name,
            absolute=True,
        )
        return header

    def get_name(self) -> Optional[str]:
        """
        Attempts to get the name of the object
        Sanity checks size members to avoid FPs
        Returns None if any issues detected
        """
        try:
            name_info = self.NameInfo.Name
            if (
                name_info.Length == 0
                or name_info.MaximumLength == 0
                or name_info.Length > name_info.MaximumLength
            ):
                return None

            return name_info.String
        except (ValueError, exceptions.InvalidAddressException):
            return None



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import csv
import datetime
import json
import logging
import random
import string
import sys
from functools import wraps
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, TypeVar, Union

from volatility3.cli import text_filter
from volatility3.framework import exceptions, interfaces, renderers
from volatility3.framework.renderers import format_hints

vollog = logging.getLogger(__name__)

try:
    CAPSTONE_PRESENT = True
    import capstone
except ImportError:
    CAPSTONE_PRESENT = False
    vollog.debug("Disassembly library capstone not found")


def hex_bytes_as_text(value: bytes, width: int = 16) -> str:
    """Renders HexBytes as text.

    Args:
        value: A series of bytes to convert to text

    Returns:
        A text representation of the hexadecimal bytes plus their ascii equivalents, separated by newline characters
    """
    if not isinstance(value, bytes):
        raise TypeError(f"hex_bytes_as_text takes bytes not: {type(value)}")

    printables = ""
    output = "\n"
    for count, byte in enumerate(value):
        output += f"{byte:02x} "
        char = chr(byte)
        printables += char if 0x20 <= byte <= 0x7E else "."
        if count % width == width - 1:
            output += printables
            if count < len(value) - 1:
                output += "\n"
            printables = ""

    # Handle leftovers when the length is not a multiple of width
    if printables:
        padding = width - len(printables)
        output += "   " * padding
        output += printables
        output += " " * padding

    return output


def multitypedata_as_text(value: format_hints.MultiTypeData) -> str:
    """Renders the bytes as a string where possible, otherwise it displays hex data

    This attempts to convert the string based on its encoding and if no data's been lost due to the split on the null character, then it displays it as is
    """
    if value.show_hex:
        return hex_bytes_as_text(value)
    string_representation = str(value, encoding=value.encoding, errors="replace")
    if value.split_nulls and (
        (len(value) / 2 - 1) <= len(string_representation) <= (len(value) / 2)
    ):
        return "\n".join(string_representation.split("\x00"))
    if (
        len(string_representation) - 1
        <= len(string_representation.split("\x00")[0])
        <= len(string_representation)
    ):
        return string_representation.split("\x00")[0]
    return hex_bytes_as_text(value)


T = TypeVar("T")


def optional(
    func: Callable[[Union[interfaces.renderers.BaseAbsentValue, T]], str],
) -> Callable[[T], str]:
    @wraps(func)
    def wrapped(x: Any) -> str:
        if isinstance(x, interfaces.renderers.BaseAbsentValue):
            if isinstance(x, renderers.NotApplicableValue):
                return "N/A"
            else:
                return "-"
        return func(x)

    return wrapped


def quoted_optional(func: Callable) -> Callable:
    @wraps(func)
    def wrapped(x: Any) -> str:
        result = optional(func)(x)
        if result == "-" or result == "N/A":
            return ""
        if isinstance(x, format_hints.MultiTypeData) and x.converted_int:
            return f"{result}"
        if isinstance(x, int) and not isinstance(
            x, (format_hints.Hex, format_hints.Bin)
        ):
            return f"{result}"
        return f'"{result}"'

    return wrapped


def display_disassembly(disasm: renderers.Disassembly) -> str:
    """Renders a disassembly renderer type into string format.

    Args:
        disasm: Input disassembly objects

    Returns:
        A string as rendered by capstone where available, otherwise output as if it were just bytes
    """

    if CAPSTONE_PRESENT:
        disasm_types = {
            "intel": capstone.Cs(capstone.CS_ARCH_X86, capstone.CS_MODE_32),
            "intel64": capstone.Cs(capstone.CS_ARCH_X86, capstone.CS_MODE_64),
            "arm": capstone.Cs(capstone.CS_ARCH_ARM, capstone.CS_MODE_ARM),
            "arm64": capstone.Cs(capstone.CS_ARCH_ARM64, capstone.CS_MODE_ARM),
        }
        output = ""
        if disasm.architecture is not None:
            for i in disasm_types[disasm.architecture].disasm(
                disasm.data, disasm.offset
            ):
                output += f"\n{i.address:#x}:\t{i.mnemonic}\t{i.op_str}"
        return output
    return QuickTextRenderer._type_renderers[bytes](disasm.data)


class CLITypeRenderer(interfaces.renderers.TypeRendererInterface):
    def __init__(self, func):
        super().__init__(func=optional(func))


class LayerDataRenderer(CLITypeRenderer):
    """Renders a LayerData object into data/bytes"""

    def __init__(self):
        self.context_byte_len = 0
        self.width = 16
        self.display_offset = False
        self.display_hex = True
        self.display_ascii = True

        def render(
            data: Union[renderers.LayerData, interfaces.renderers.BaseAbsentValue],
        ) -> str:
            if isinstance(data, interfaces.renderers.BaseAbsentValue):
                # FIXME: Do something cleverer here
                return ""

            specific_data, error_bytes = self.render_bytes(data)

            printables = ""
            output = "\n"
            for count, byte in enumerate(specific_data):
                if count not in error_bytes:
                    output += f"{byte:02x} "
                    char = chr(byte)
                    printables += char if 0x20 <= byte <= 0x7E else "."
                else:
                    output += "__ "
                    printables += "."
                if count % self.width == self.width - 1:
                    output += printables
                    if count < len(specific_data) - 1:
                        output += "\n"
                    printables = ""

            # Handle leftovers when the length is not a multiple of width
            if printables:
                padding = self.width - len(printables)
                output += "   " * padding
                output += printables
                output += " " * padding

            return output

        render_func = render
        return super().__init__(render_func)

    def render_bytes(self, data: renderers.LayerData) -> Tuple[bytes, Set[int]]:
        """Renders a valid LayerData into bytes (with context bytes)"""
        context_byte_len = self.context_byte_len if not data.no_surrounding else 0

        layer = data.context.layers[data.layer_name]
        # Map of the holes
        error_bytes = set()
        start_offset = data.offset - context_byte_len
        end_offset = data.offset + data.length + context_byte_len
        if isinstance(layer, interfaces.layers.TranslationLayerInterface):
            error_bytes = set()
            mapping = iter(layer.mapping(start_offset, end_offset, True))
            current_map = next(mapping)
            for i in range(start_offset, end_offset):
                # Run through the bytes, check if they're present
                offset, sublength, _, _, _ = current_map
                if i < offset:
                    error_bytes.add(i - start_offset)
                if i > offset + sublength:
                    try:
                        current_map = next(mapping)
                    except StopIteration:
                        pass
                    offset, sublength, _, _, _ = current_map
                if i > offset + sublength:
                    error_bytes.add(i - start_offset)

        # Padded data
        specific_data = data.context.layers[data.layer_name].read(
            start_offset,
            end_offset - start_offset,
            True,
        )

        return specific_data, error_bytes


class CLIRenderer(interfaces.renderers.Renderer):
    """Class to add specific requirements for CLI renderers."""

    _type_renderers = {
        format_hints.Bin: CLITypeRenderer(lambda x: f"0b{x:b}"),
        format_hints.Hex: CLITypeRenderer(lambda x: f"0x{x:x}"),
        format_hints.HexBytes: CLITypeRenderer(hex_bytes_as_text),
        format_hints.MultiTypeData: CLITypeRenderer(multitypedata_as_text),
        renderers.Disassembly: CLITypeRenderer(display_disassembly),
        bytes: CLITypeRenderer(lambda x: " ".join(f"{b:02x}" for b in x)),
        renderers.LayerData: LayerDataRenderer(),
        datetime.datetime: CLITypeRenderer(
            lambda x: x.strftime("%Y-%m-%d %H:%M:%S.%f %Z")
        ),
        "default": CLITypeRenderer(lambda x: f"{x}"),
    }

    name = "unnamed"
    structured_output = False
    filter: Optional[text_filter.CLIFilter] = None
    column_hide_list: Optional[list] = None

    def ignored_columns(
        self,
        grid: interfaces.renderers.TreeGrid,
    ) -> List[interfaces.renderers.Column]:
        ignored_column_list = []
        if self.column_hide_list:
            for column in grid.columns:
                accept = True
                for column_prefix in self.column_hide_list:
                    if column.name.lower().startswith(column_prefix.lower()):
                        accept = False
                if not accept:
                    ignored_column_list.append(column)
        elif self.column_hide_list is None:
            return []

        if len(ignored_column_list) == len(grid.columns):
            raise exceptions.RenderException("No visible columns to render")
        vollog.info(
            f"Hiding columns: {[column.name for column in ignored_column_list]}"
        )
        return ignored_column_list


class QuickTextRenderer(CLIRenderer):
    name = "quick"

    def get_render_options(self):
        return []

    def render(self, grid: interfaces.renderers.TreeGrid) -> None:
        """Renders each column immediately to stdout.

        This does not format each line's width appropriately, it merely tab separates each field

        Args:
            grid: The TreeGrid object to render
        """
        # TODO: Docstrings
        # TODO: Improve text output
        outfd = sys.stdout

        line = []
        ignore_columns = self.ignored_columns(grid)
        for column in grid.columns:
            # Ignore the type because namedtuples don't realize they have accessible attributes
            if column not in ignore_columns:
                line.append(f"{column.name}")
        outfd.write("\n{}\n".format("\t".join(line)))

        def visitor(node: interfaces.renderers.TreeNode, accumulator):
            line = []
            for column_index, column in enumerate(grid.columns):
                renderer = self._type_renderers.get(
                    column.type, self._type_renderers["default"]
                )
                if column not in ignore_columns:
                    line.append(renderer(node.values[column_index]))

            if self.filter and self.filter.filter(line):
                return accumulator

            accumulator.write("\n")
            # Nodes always have a path value, giving them a path_depth of at least 1, we use max just in case
            accumulator.write(
                "*" * max(0, node.path_depth - 1)
                + ("" if (node.path_depth <= 1) else " ")
            )
            accumulator.write("{}".format("\t".join(line)))
            accumulator.flush()
            return accumulator

        if not grid.populated:
            grid.populate(visitor, outfd)
        else:
            grid.visit(node=None, function=visitor, initial_accumulator=outfd)

        outfd.write("\n")


class NoneRenderer(CLIRenderer):
    """Outputs no results"""

    name = "none"

    def get_render_options(self):
        return []

    def render(self, grid: interfaces.renderers.TreeGrid) -> None:
        if not grid.populated:
            grid.populate(lambda x, y: True, True)


class CSVRenderer(CLIRenderer):
    name = "csv"
    structured_output = True

    def get_render_options(self):
        return []

    def render(self, grid: interfaces.renderers.TreeGrid) -> None:
        """Renders each row immediately to stdout.

        Args:
            grid: The TreeGrid object to render
        """
        outfd = sys.stdout
        ignore_columns = self.ignored_columns(grid)

        header_list = ["TreeDepth"]
        for column in grid.columns:
            # Ignore the type because namedtuples don't realize they have accessible attributes
            if column not in ignore_columns:
                header_list.append(f"{column.name}")

        writer = csv.DictWriter(
            outfd, header_list, lineterminator="\n", escapechar="\\"
        )
        writer.writeheader()

        def visitor(node: interfaces.renderers.TreeNode, accumulator):
            # Nodes always have a path value, giving them a path_depth of at least 1, we use max just in case
            row = {"TreeDepth": str(max(0, node.path_depth - 1))}
            line = []
            for column_index, column in enumerate(grid.columns):
                renderer = self._type_renderers.get(
                    column.type, self._type_renderers["default"]
                )
                row[f"{column.name}"] = renderer(node.values[column_index])
                if column not in ignore_columns:
                    line.append(row[f"{column.name}"])
                else:
                    del row[f"{column.name}"]

            if self.filter and self.filter.filter(line):
                return accumulator

            accumulator.writerow(row)
            return accumulator

        if not grid.populated:
            grid.populate(visitor, writer)
        else:
            grid.visit(node=None, function=visitor, initial_accumulator=writer)

        outfd.write("\n")


class PrettyTextRenderer(CLIRenderer):
    name = "pretty"

    def get_render_options(self):
        return []

    def render(self, grid: interfaces.renderers.TreeGrid) -> None:
        """Renders each column immediately to stdout.

        This does not format each line's width appropriately, it merely tab separates each field

        Args:
            grid: The TreeGrid object to render
        """
        # TODO: Docstrings
        # TODO: Improve text output
        outfd = sys.stdout

        sys.stderr.write("Formatting...\n")

        ignore_columns = self.ignored_columns(grid)
        display_alignment = ">"
        column_separator = " | "

        tree_indent_column = "".join(
            random.choices(string.ascii_uppercase + string.digits, k=20)
        )
        max_column_widths = dict(
            [(column.name, len(column.name)) for column in grid.columns]
        )

        def visitor(
            node: interfaces.renderers.TreeNode,
            accumulator: List[Tuple[int, Dict[interfaces.renderers.Column, bytes]]],
        ) -> List[Tuple[int, Dict[interfaces.renderers.Column, bytes]]]:
            # Nodes always have a path value, giving them a path_depth of at least 1, we use max just in case
            max_column_widths[tree_indent_column] = max(
                max_column_widths.get(tree_indent_column, 0), node.path_depth
            )

            line = {}
            rendered_line = []
            for column_index, column in enumerate(grid.columns):
                renderer = self._type_renderers.get(
                    column.type, self._type_renderers["default"]
                )
                data = renderer(node.values[column_index])
                field_width = max(
                    [len(self.tab_stop(x)) for x in f"{data}".split("\n")]
                )
                max_column_widths[column.name] = max(
                    max_column_widths.get(column.name, len(column.name)), field_width
                )
                if column not in ignore_columns:
                    line[column] = data.split("\n")
                rendered_line.append(data)

            if self.filter and self.filter.filter(rendered_line):
                return accumulator

            accumulator.append((node.path_depth, line))
            return accumulator

        final_output: List[
            Tuple[int, Dict[interfaces.renderers.Column, list[str]]]
        ] = []
        if not grid.populated:
            grid.populate(visitor, final_output)
        else:
            grid.visit(node=None, function=visitor, initial_accumulator=final_output)

        # Always align the tree to the left
        format_string_list = [
            "{0:<" + str(max_column_widths.get(tree_indent_column, 0)) + "s}"
        ]
        column_offset = 0
        for column_index, column in enumerate(grid.columns):
            if column not in ignore_columns:
                format_string_list.append(
                    "{"
                    + str(column_index - column_offset + 1)
                    + ":"
                    + display_alignment
                    + str(max_column_widths[column.name])
                    + "s}"
                )
            else:
                column_offset += 1

        format_string = column_separator.join(format_string_list) + "\n"

        column_titles = [""] + [
            column.name for column in grid.columns if column not in ignore_columns
        ]

        outfd.write(format_string.format(*column_titles))
        for depth, line in final_output:
            nums_line = max([len(line[column]) for column in line])
            for column in line:
                if column in ignore_columns:
                    del line[column]
                else:
                    line[column] = line[column] + (
                        [""] * (nums_line - len(line[column]))
                    )
            for index in range(nums_line):
                if index == 0:
                    outfd.write(
                        format_string.format(
                            "*" * depth,
                            *[self.tab_stop(line[column][index]) for column in line],
                        )
                    )
                else:
                    outfd.write(
                        format_string.format(
                            " " * depth,
                            *[self.tab_stop(line[column][index]) for column in line],
                        )
                    )

    def tab_stop(self, line: str) -> str:
        tab_width = 8
        while line.find("\t") >= 0:
            i = line.find("\t")
            pad = " " * (tab_width - (i % tab_width))
            line = line.replace("\t", pad, 1)
        return line


class JsonRenderer(CLIRenderer):
    _type_renderers = {
        format_hints.HexBytes: lambda x: (
            x.hex(" ")
            if not isinstance(x, interfaces.renderers.BaseAbsentValue)
            else "N/A"
        ),
        renderers.Disassembly: quoted_optional(display_disassembly),
        format_hints.MultiTypeData: quoted_optional(multitypedata_as_text),
        renderers.LayerData: lambda x: (
            LayerDataRenderer().render_bytes(x)[0].hex(" ")
            if not isinstance(x, interfaces.renderers.BaseAbsentValue)
            else "N/A"
        ),
        bytes: optional(lambda x: " ".join(f"{b:02x}" for b in x)),
        datetime.datetime: lambda x: (
            x.isoformat()
            if not isinstance(x, interfaces.renderers.BaseAbsentValue)
            else None
        ),
        "default": lambda x: x,
    }

    name = "JSON"
    structured_output = True

    def get_render_options(self) -> List[interfaces.renderers.RenderOption]:
        return []

    def output_result(self, outfd, result):
        """Outputs the JSON data to a file in a particular format"""
        outfd.write(f"{json.dumps(result, indent=2, sort_keys=True)}\n")

    def render(self, grid: interfaces.renderers.TreeGrid):
        outfd = sys.stdout

        outfd.write("\n")
        final_output: Tuple[
            Dict[str, List[interfaces.renderers.TreeNode]],
            List[interfaces.renderers.TreeNode],
        ] = ({}, [])

        ignore_columns = self.ignored_columns(grid)

        def visitor(
            node: interfaces.renderers.TreeNode,
            accumulator: Tuple[Dict[str, Dict[str, Any]], List[Dict[str, Any]]],
        ) -> Tuple[Dict[str, Dict[str, Any]], List[Dict[str, Any]]]:
            # Nodes always have a path value, giving them a path_depth of at least 1, we use max just in case
            acc_map, final_tree = accumulator
            node_dict: Dict[str, Any] = {"__children": []}
            line = []
            for column_index, column in enumerate(grid.columns):
                if column in ignore_columns:
                    continue
                renderer = self._type_renderers.get(
                    column.type, self._type_renderers["default"]
                )
                data = renderer(list(node.values)[column_index])
                if isinstance(data, interfaces.renderers.BaseAbsentValue):
                    data = None
                node_dict[column.name] = data
                line.append(data)

            if self.filter and self.filter.filter(line):
                return accumulator

            # Only add if the parent hasn't been filtered out
            if node.parent and node.parent.path in acc_map:
                acc_map[node.parent.path]["__children"].append(node_dict)
            else:
                final_tree.append(node_dict)
            acc_map[node.path] = node_dict

            return (acc_map, final_tree)

        if not grid.populated:
            grid.populate(visitor, final_output)
        else:
            grid.visit(node=None, function=visitor, initial_accumulator=final_output)

        self.output_result(outfd, final_output[1])


class JsonLinesRenderer(JsonRenderer):
    name = "JSONL"

    def output_result(self, outfd, result):
        """Outputs the JSON results as JSON lines"""
        for line in result:
            outfd.write(json.dumps(line, sort_keys=True))
            outfd.write("\n")



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
from typing import Dict, Optional, Tuple

from volatility3.framework import constants, exceptions, interfaces, objects
from volatility3.framework.constants import linux as linux_constants

vollog = logging.getLogger(__name__)


class elf(objects.StructType):
    """
    Class used to create elf objects. It overrides the typename to `Elf32_` or `Elf64_`,
    depending on the corresponding value on e_ident
    """

    def __init__(
        self,
        context: interfaces.context.ContextInterface,
        type_name: str,
        object_info: interfaces.objects.ObjectInformation,
        size: int,
        members: Dict[str, Tuple[int, interfaces.objects.Template]],
    ) -> None:
        super().__init__(
            context=context,
            type_name=type_name,
            object_info=object_info,
            size=size,
            members=members,
        )

        layer_name = self.vol.layer_name
        symbol_table_name = self.get_symbol_table_name()
        # We read the MAGIC: (0x0 to 0x4) 0x7f 0x45 0x4c 0x46
        try:
            magic = self._context.object(
                symbol_table_name + constants.BANG + "unsigned long",
                layer_name=layer_name,
                offset=object_info.offset,
            )
        except (
            exceptions.PagedInvalidAddressException,
            exceptions.InvalidAddressException,
        ) as excp:
            vollog.debug(
                f"Unable to check magic bytes for ELF file at offset {hex(object_info.offset)} in layer {layer_name}: {excp}"
            )
            return None

        # Check validity
        if magic != 0x464C457F:  # e.g. ELF
            return None

        # We need to read the EI_CLASS (0x4 offset)
        ei_class = self._context.object(
            symbol_table_name + constants.BANG + "unsigned char",
            layer_name=layer_name,
            offset=object_info.offset + linux_constants.ELF_IDENT.EI_CLASS,
        )

        if ei_class == linux_constants.ELF_CLASS.ELFCLASS32:
            self._type_prefix = "Elf32_"
            self._ei_class_size = 32
        elif ei_class == linux_constants.ELF_CLASS.ELFCLASS64:
            self._type_prefix = "Elf64_"
            self._ei_class_size = 64
        else:
            raise ValueError(f"Unsupported ei_class value {ei_class}")

        # Construct the full header
        self._hdr = self._context.object(
            symbol_table_name + constants.BANG + self._type_prefix + "Ehdr",
            layer_name=layer_name,
            offset=object_info.offset,
        )
        self._offset = object_info.offset

        self._cached_symtab = None
        self._cached_strtab = None

    def is_valid(self):
        """
        Determine whether it is a valid object
        """
        if hasattr(self, "_type_prefix") and hasattr(self, "_hdr"):
            return self._type_prefix is not None and self._hdr is not None
        else:
            return False

    def __getattr__(self, name):
        # Just redirect to the corresponding header
        if name[0:2] == "e_" and name in dir(self._hdr):
            return self._hdr.__getattr__(name)
        else:
            return self.__getattribute__(name)

    def __dir__(self):
        return self._hdr.__dir__() + [
            "get_program_headers",
            "is_valid",
            "get_section_headers",
            "get_symbols",
            "__dir__",
        ]

    def get_program_headers(self):
        program_headers = self._context.object(
            self.get_symbol_table_name() + constants.BANG + "array",
            layer_name=self.vol.layer_name,
            offset=self._offset + self.e_phoff,
            subtype=self._context.symbol_space.get_type(
                self.get_symbol_table_name()
                + constants.BANG
                + self._type_prefix
                + "Phdr"
            ),
            count=self.e_phnum,
        )

        for prog_header in program_headers:
            prog_header.parent_e_type = self.e_type
            prog_header.parent_offset = self._offset
            prog_header.type_prefix = self._type_prefix
            yield prog_header

    def get_section_headers(self):
        section_headers = self._context.object(
            self.get_symbol_table_name() + constants.BANG + "array",
            layer_name=self.vol.layer_name,
            offset=self._offset + self.e_shoff,
            subtype=self._context.symbol_space.get_type(
                self.get_symbol_table_name()
                + constants.BANG
                + self._type_prefix
                + "Shdr"
            ),
            count=self.e_shnum,
        )
        return section_headers

    def get_link_maps(self, kernel_symbol_table_name):
        """Get the ELF link map objects for the given VMA address

        Args:
            kernel_symbol_table_name (str): Kernel symbol table name

        Yields:
            The ELF link map objects
        """
        got_entry_size = self._ei_class_size // 8

        elf_symbol_table = self.get_symbol_table_name()

        link_maps_seen = set()
        for phdr in self.get_program_headers():
            try:
                if phdr.p_type.description != "PT_DYNAMIC":
                    continue
            except ValueError:
                vollog.log(
                    constants.LOGLEVEL_VVVV,
                    f"Skipping unknown ELF program header type: {phdr.p_type}",
                )
                continue

            for dsec in phdr.dynamic_sections():
                try:
                    if dsec.d_tag.description != "DT_PLTGOT":
                        continue
                except ValueError:
                    vollog.log(
                        constants.LOGLEVEL_VVVV,
                        f"Skipping unknown ELF dynamic section type: {dsec.d_tag}",
                    )
                    continue

                got_start = dsec.d_ptr

                # link_map is stored at the second GOT entry
                link_map_addr = got_start + got_entry_size

                # It needs the kernel symbol table to create a pointer
                link_map_ptr = self._context.object(
                    kernel_symbol_table_name + constants.BANG + "pointer",
                    offset=link_map_addr,
                    layer_name=self.vol.layer_name,
                )
                if not link_map_ptr:
                    vollog.log(
                        constants.LOGLEVEL_VVVV,
                        f"Invalid ELF link map pointer at 0x{link_map_addr:x}",
                    )
                    continue

                linkmap_symname = (
                    elf_symbol_table + constants.BANG + self._type_prefix + "LinkMap"
                )
                try:
                    link_map = self._context.object(
                        object_type=linkmap_symname,
                        offset=link_map_ptr,
                        layer_name=self.vol.layer_name,
                    )
                except exceptions.InvalidAddressException:
                    vollog.log(
                        constants.LOGLEVEL_VVVV,
                        f"Invalid ELF link map address at 0x{link_map_ptr:x}",
                    )
                    continue

                while link_map and link_map.vol.offset != 0:
                    if link_map.vol.offset in link_maps_seen:
                        break
                    link_maps_seen.add(link_map.vol.offset)

                    yield link_map

                    try:
                        link_map = self._context.object(
                            object_type=linkmap_symname,
                            offset=link_map.l_next,
                            layer_name=self.vol.layer_name,
                        )
                    except exceptions.InvalidAddressException:
                        vollog.log(
                            constants.LOGLEVEL_VVVV,
                            f"ELF link map linked list is corrupt at 0x{self.vol.offset:x}",
                        )
                        break

    def _find_symbols(self):
        dt_strtab = None
        dt_symtab = None
        dt_strent = None

        for phdr in self.get_program_headers():
            # Find PT_DYNAMIC segment
            try:
                if phdr.p_type.description != "PT_DYNAMIC":
                    continue
            except ValueError:
                vollog.log(
                    constants.LOGLEVEL_VVVV,
                    f"Skipping unknown ELF program header type: {phdr.p_type}",
                )
                continue

            # This section contains pointers to the strtab, symtab, and strent sections
            for dsec in phdr.dynamic_sections():
                try:
                    dtag = dsec.d_tag.description
                except ValueError:
                    vollog.log(
                        constants.LOGLEVEL_VVVV,
                        f"Skipping unknown ELF dynamic section type: {dsec.d_tag}",
                    )
                    continue

                if dtag == "DT_STRTAB":
                    dt_strtab = dsec.d_ptr

                elif dtag == "DT_SYMTAB":
                    dt_symtab = dsec.d_ptr

                elif dtag == "DT_SYMENT":
                    # Size of the symtab symbol entry
                    dt_strent = dsec.d_ptr

            break

        if not (dt_strtab and dt_symtab and dt_strent):
            return None

        self._cached_symtab = dt_symtab
        self._cached_strtab = dt_strtab

        # Calculate number of symbol entries assuming that strtab follows symtab
        if dt_symtab < dt_strtab:
            self._cached_numsyms = (dt_strtab - dt_symtab) // dt_strent
        else:
            self._cached_numsyms = 1024

    def get_symbols(self):
        if self._cached_symtab is None:
            self._find_symbols()

        if self._cached_symtab is None:
            return None

        symtab_arr = self._context.object(
            self.get_symbol_table_name() + constants.BANG + "array",
            layer_name=self.vol.layer_name,
            offset=self._cached_symtab,
            subtype=self._context.symbol_space.get_type(
                self.get_symbol_table_name()
                + constants.BANG
                + self._type_prefix
                + "Sym"
            ),
            count=self._cached_numsyms,
        )

        for sym in symtab_arr:
            sym.cached_strtab = self._cached_strtab
            yield sym


class elf_sym(objects.StructType):
    """An elf symbol entry"""

    _MAX_NAME_LENGTH = linux_constants.KSYM_NAME_LEN

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._cached_strtab = None

    @property
    def cached_strtab(self):
        return self._cached_strtab

    @cached_strtab.setter
    def cached_strtab(self, cached_strtab):
        self._cached_strtab = cached_strtab

    def get_name(self) -> Optional[str]:
        """Returns the symbol name"""

        try:
            addr = self._cached_strtab + self.st_name
        except exceptions.InvalidAddressException:
            return None

        layer = self._context.layers[self.vol.layer_name]
        name_bytes = layer.read(addr, self._MAX_NAME_LENGTH, pad=True)
        if not name_bytes:
            return None

        idx = name_bytes.find(b"\x00")
        if idx != -1:
            name_bytes = name_bytes[:idx]

        return name_bytes.decode("utf-8", errors="replace")


class elf_phdr(objects.StructType):
    """An elf program header"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._parent_e_type = None
        self._parent_offset = None
        self._type_prefix = None

    @property
    def parent_e_type(self):
        return self._parent_e_type

    @parent_e_type.setter
    def parent_e_type(self, e_type):
        self._parent_e_type = e_type

    @property
    def parent_offset(self):
        return self._parent_offset

    @parent_offset.setter
    def parent_offset(self, offset):
        self._parent_offset = offset

    @property
    def type_prefix(self):
        return self._type_prefix

    @type_prefix.setter
    def type_prefix(self, prefix):
        self._type_prefix = prefix

    def get_vaddr(self):
        offset = self.__getattr__("p_vaddr")

        try:
            if self._parent_e_type.description == "ET_DYN":
                offset = self._parent_offset + offset
        except ValueError:
            # Unknown ELF object file type. Anyway, if the ELF object file type is not a
            # shared object (ET_DYN), the virtual address is 'p_vaddr'.
            vollog.log(
                constants.LOGLEVEL_VVVV,
                f"Skipping unknown ELF object type: {self._parent_e_type}",
            )

        return offset

    def dynamic_sections(self):
        # sanity check
        try:
            if self.p_type.description != "PT_DYNAMIC":
                return None
        except ValueError:
            # If the value is outside the ones declared in the enumeration, an
            # exception is raised
            vollog.log(
                constants.LOGLEVEL_VVVV,
                f"Skipping unknown ELF program header type: {self.p_type}",
            )
            return None

        # the buffer of array starts at elf_base + our virtual address ( offset )
        arr_start = self.get_vaddr()

        symbol_table_name = self.get_symbol_table_name()

        rtsize = self._context.symbol_space.get_type(
            symbol_table_name + constants.BANG + self._type_prefix + "Dyn"
        ).size

        for i in range(256):
            # use the real size
            idx = i * rtsize

            dyn = self._context.object(
                symbol_table_name + constants.BANG + self._type_prefix + "Dyn",
                layer_name=self.vol.layer_name,
                offset=arr_start + idx,
            )

            yield dyn

            if dyn.d_tag == 0:
                break


class elf_linkmap(objects.StructType):
    def get_name(self):
        try:
            buf = self._context.layers.read(self.vol.layer_name, self.l_name, 256)
        except exceptions.InvalidAddressException:
            # Protection against memory smear
            vollog.log(
                constants.LOGLEVEL_VVVV,
                f"Invalid l_name address for ELF link map at 0x{self.vol.offset:x}",
            )
            return None

        idx = buf.find(b"\x00")
        if idx != -1:
            buf = buf[:idx]
        return buf.decode("utf-8", errors="replace")


class_types = {
    "Elf": elf,
    "Elf64_Phdr": elf_phdr,
    "Elf32_Phdr": elf_phdr,
    "Elf32_Sym": elf_sym,
    "Elf64_Sym": elf_sym,
    "Elf32_LinkMap": elf_linkmap,
    "Elf64_LinkMap": elf_linkmap,
}



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
"""A module containing a collection of plugins that produce data typically
found in mac's /proc file system."""

import datetime
import struct

from volatility3.framework import constants, renderers, symbols
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins
from volatility3.framework.layers import scanners
from volatility3.framework.objects import utility
from volatility3.framework.symbols.linux import bash
from volatility3.plugins import timeliner
from volatility3.plugins.mac import pslist


class Bash(plugins.PluginInterface, timeliner.TimeLinerInterface):
    """Recovers bash command history from memory."""

    _required_framework_version = (2, 0, 0)

    @classmethod
    def get_requirements(cls):
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Kernel module for the OS",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.VersionRequirement(
                name="timeliner",
                component=timeliner.TimeLinerInterface,
                version=(1, 0, 0),
            ),
            requirements.VersionRequirement(
                name="multi_string_scanner",
                component=scanners.MultiStringScanner,
                version=(1, 0, 0),
            ),
            requirements.VersionRequirement(
                name="bytes_scanner",
                component=scanners.BytesScanner,
                version=(1, 0, 0),
            ),
            requirements.ListRequirement(
                name="pid",
                description="Filter on specific process IDs",
                element_type=int,
                optional=True,
            ),
        ]

    def _generator(self, tasks):
        darwin = self.context.modules[self.config["kernel"]]
        is_32bit = not symbols.symbol_table_is_64bit(
            context=self.context, symbol_table_name=darwin.symbol_table_name
        )
        if is_32bit:
            pack_format = "I"
            bash_json_file = "bash32"
        else:
            pack_format = "Q"
            bash_json_file = "bash64"

        bash_table_name = bash.BashIntermedSymbols.create(
            self.context, self.config_path, "linux", bash_json_file
        )

        ts_offset = self.context.symbol_space.get_type(
            bash_table_name + constants.BANG + "hist_entry"
        ).relative_child_offset("timestamp")

        for task in tasks:
            task_name = utility.array_to_string(task.p_comm)
            if task_name not in ["bash", "sh", "dash"]:
                continue

            proc_layer_name = task.add_process_layer()
            if proc_layer_name is None:
                continue

            proc_layer = self.context.layers[proc_layer_name]

            bang_addrs = []

            # find '#' values on the heap
            for address in proc_layer.scan(
                self.context,
                scanners.BytesScanner(b"#"),
                sections=task.get_process_memory_sections(
                    self.context, self.config["kernel"], rw_no_file=True
                ),
            ):
                bang_addrs.append(struct.pack(pack_format, address))

            history_entries = []

            for address, _ in proc_layer.scan(
                self.context,
                scanners.MultiStringScanner(bang_addrs),
                sections=task.get_process_memory_sections(
                    self.context, self.config["kernel"], rw_no_file=True
                ),
            ):
                hist = self.context.object(
                    bash_table_name + constants.BANG + "hist_entry",
                    offset=address - ts_offset,
                    layer_name=proc_layer_name,
                )

                if hist.is_valid():
                    history_entries.append(hist)

            for hist in sorted(history_entries, key=lambda x: x.get_time_as_integer()):
                yield (
                    0,
                    (
                        int(task.p_pid),
                        task_name,
                        hist.get_time_object(),
                        hist.get_command(),
                    ),
                )

    def run(self):
        filter_func = pslist.PsList.create_pid_filter(self.config.get("pid", None))
        list_tasks = pslist.PsList.get_list_tasks(
            self.config.get("pslist_method", pslist.PsList.pslist_methods[0])
        )

        return renderers.TreeGrid(
            [
                ("PID", int),
                ("Process", str),
                ("CommandTime", datetime.datetime),
                ("Command", str),
            ],
            self._generator(
                list_tasks(self.context, self.config["kernel"], filter_func=filter_func)
            ),
        )

    def generate_timeline(self):
        filter_func = pslist.PsList.create_pid_filter(self.config.get("pid", None))
        list_tasks = pslist.PsList.get_list_tasks(
            self.config.get("pslist_method", pslist.PsList.pslist_methods[0])
        )

        for row in self._generator(
            list_tasks(self.context, self.config["kernel"], filter_func=filter_func)
        ):
            _depth, row_data = row
            description = f'{row_data[0]} ({row_data[1]}): "{row_data[3]}"'
            yield (description, timeliner.TimeLinerType.CREATED, row_data[2])



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
from typing import Any, Dict, Iterable, List, Tuple

from volatility3.framework import interfaces, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins
from volatility3.framework.layers import resources
from volatility3.framework.renderers import format_hints

vollog = logging.getLogger(__name__)

USE_YARA_X = False

try:
    import yara_x

    USE_YARA_X = True

except ImportError:
    try:
        import yara

        if tuple(int(x) for x in yara.__version__.split(".")) < (3, 8):
            raise ImportError

        vollog.debug("Using yara-python module")

    except ImportError:
        vollog.info(
            "Neither yara-x nor yara-python (>3.8.0) module was found, plugin (and dependent plugins) not available"
        )
        raise


class YaraScanner(interfaces.layers.ScannerInterface):
    _version = (2, 1, 1)

    # yara.Rules isn't exposed, so we can't type this properly
    def __init__(self, rules) -> None:
        super().__init__()
        if rules is None:
            raise ValueError("No rules provided to YaraScanner")
        self._rules = rules
        self.st_object = (
            None
            if USE_YARA_X
            else not tuple(int(x) for x in yara.__version__.split(".")) < (4, 3)
        )

    def __call__(
        self, data: bytes, data_offset: int
    ) -> Iterable[Tuple[int, str, str, bytes]]:
        if USE_YARA_X:
            for match in self._rules.scan(data).matching_rules:
                for match_string in match.patterns:
                    for instance in match_string.matches:
                        yield (
                            instance.offset + data_offset,
                            f"{match.namespace}.{match.identifier}",
                            match_string.identifier,
                            data[instance.offset : instance.offset + instance.length],
                        )
        else:
            for match in self._rules.match(data=data):
                if YaraScan.yara_returns_instances():
                    for match_string in match.strings:
                        for instance in match_string.instances:
                            yield (
                                instance.offset + data_offset,
                                match.rule,
                                match_string.identifier,
                                instance.matched_data,
                            )
                else:
                    for offset, name, value in match.strings:
                        yield (offset + data_offset, match.rule, name, value)

    @classmethod
    def get_rule(cls, rule):
        if USE_YARA_X:
            return yara_x.compile(f"rule r1 {{strings: $a = {rule} condition: $a}}")
        return yara.compile(
            sources={"n": f"rule r1 {{strings: $a = {rule} condition: $a}}"}
        )

    @classmethod
    def from_compiled_file(cls, filepath):
        with resources.ResourceAccessor().open(filepath, "rb") as fp:
            if USE_YARA_X:
                return yara_x.Rules.deserialize_from(file=fp)
            return yara.load(file=fp)

    @classmethod
    def from_file(cls, filepath):
        with resources.ResourceAccessor().open(filepath, "rb") as fp:
            if USE_YARA_X:
                return yara_x.compile(fp.read().decode())
            return yara.compile(file=fp)


class YaraScan(plugins.PluginInterface):
    """Scans kernel memory using yara rules (string or file)."""

    _required_framework_version = (2, 22, 0)
    _version = (2, 0, 1)
    _yara_x = USE_YARA_X

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        """Returns the requirements needed to run yarascan directly, combining the TranslationLayerRequirement
        and the requirements from get_yarascan_option_requirements."""
        return cls.get_yarascan_option_requirements() + [
            requirements.TranslationLayerRequirement(
                name="primary",
                description="Memory layer for the kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="yarascanner",
                component=YaraScanner,
                version=(2, 1, 1),
            ),
        ]

    @classmethod
    def get_yarascan_option_requirements(
        cls,
    ) -> List[interfaces.configuration.RequirementInterface]:
        """Returns the requirements needed for the command lines options used by yarascan. This can
        then also be used by other plugins that are using yarascan. This does not include a
        TranslationLayerRequirement or a ModuleRequirement."""
        return [
            requirements.BooleanRequirement(
                name="insensitive",
                description="Makes the search case insensitive",
                default=False,
                optional=True,
            ),
            requirements.BooleanRequirement(
                name="wide",
                description="Match wide (unicode) strings",
                default=False,
                optional=True,
            ),
            requirements.StringRequirement(
                name="yara_string",
                description="Yara rules (as a string)",
                optional=True,
            ),
            requirements.URIRequirement(
                name="yara_file",
                description="Yara rules (as a file)",
                optional=True,
            ),
            # This additional requirement is to follow suit with upstream, who feel that compiled rules could potentially be used to execute malicious code
            # As such, there's a separate option to run compiled files, as happened with yara-3.9 and later
            requirements.URIRequirement(
                name="yara_compiled_file",
                description="Yara compiled rules (as a file)",
                optional=True,
            ),
            requirements.IntRequirement(
                name="max_size",
                default=0x40000000,
                description="Set the maximum size (default is 1GB)",
                optional=True,
            ),
        ]

    @classmethod
    def yara_returns_instances(cls) -> bool:
        return not tuple(int(x) for x in yara.__version__.split(".")) < (4, 3)

    @classmethod
    def process_yara_options(cls, config: Dict[str, Any]):
        rules = None
        if config.get("yara_string") is not None:
            rule = config["yara_string"]
            if rule[0] not in ["{", "/"]:
                rule = f'"{rule}"'
            if config.get("insensitive", False):
                rule += " nocase"
            if config.get("wide", False):
                rule += " wide ascii"
            rules = YaraScanner.get_rule(rule)
        elif config.get("yara_file") is not None:
            vollog.debug(f"Plain file: {config['yara_file']} - yara-x: {USE_YARA_X}")
            rules = YaraScanner.from_file(config["yara_file"])
        elif config.get("yara_compiled_file") is not None:
            vollog.debug(
                f"Compiled file: {config['yara_compiled_file']} - yara-x: {USE_YARA_X}"
            )
            rules = YaraScanner.from_compiled_file(config["yara_compiled_file"])
        else:
            vollog.error("No yara rules, nor yara rules file were specified")
        return rules

    def _generator(self):
        rules = self.process_yara_options(dict(self.config))

        layer = self.context.layers[self.config["primary"]]
        for offset, rule_name, name, value in layer.scan(
            context=self.context, scanner=YaraScanner(rules=rules)
        ):
            layer_data = renderers.LayerData(
                context=self.context,
                offset=offset,
                layer_name=layer.name,
                length=len(value),
            )
            yield 0, (format_hints.Hex(offset), rule_name, name, layer_data)

    def run(self):
        return renderers.TreeGrid(
            [
                ("Offset", format_hints.Hex),
                ("Rule", str),
                ("Component", str),
                ("Value", renderers.LayerData),
            ],
            self._generator(),
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import copy
from typing import Any, Dict, Iterable, Optional, Type

from volatility3.framework import constants, interfaces, objects


class NativeTable(interfaces.symbols.NativeTableInterface):
    """Symbol List that handles Native types."""

    # FIXME: typing the native_dictionary as Tuple[interfaces.objects.ObjectInterface, str] throws many errors
    def __init__(self, name: str, native_dictionary: Dict[str, Any]) -> None:
        super().__init__(name, self)
        self._native_dictionary = copy.deepcopy(native_dictionary)
        self._overrides: Dict[str, interfaces.objects.ObjectInterface] = {}
        for native_type in self._native_dictionary:
            native_class, _native_struct = self._native_dictionary[native_type]
            self._overrides[native_type] = native_class
        # Create this once early, because it may get used a lot
        self._types = set(self._native_dictionary).union(
            {"enum", "array", "bitfield", "void", "string", "bytes", "function"}
        )

    def get_type_class(self, name: str) -> Type[interfaces.objects.ObjectInterface]:
        ntype, _ = self._native_dictionary.get(name, (objects.Integer, None))
        return ntype

    @property
    def types(self) -> Iterable[str]:
        """Returns an iterable (set) of the available symbol type names."""
        return self._types

    def get_type(self, type_name: str) -> interfaces.objects.Template:
        """Resolves a symbol name into an object template.

        This always construct a new python object, rather than using a
        cached value otherwise changes made later may affect the cached
        copy.  Calling clone after every native type construction was
        extremely slow.
        """
        # NOTE: These need updating whenever the object init signatures change
        prefix = ""
        if constants.BANG in type_name:
            name_split = type_name.split(constants.BANG)
            if len(name_split) > 2:
                raise ValueError(
                    f"SymbolName cannot contain multiple {constants.BANG} separators"
                )
            table_name, type_name = name_split
            prefix = table_name + constants.BANG

        additional: Dict[str, Any] = {}
        obj: Optional[Type[interfaces.objects.ObjectInterface]] = None
        if type_name == "void" or type_name == "function":
            obj = objects.Void
        elif type_name == "array":
            obj = objects.Array
            additional = {"count": 0, "subtype": self.get_type("void")}
        elif type_name == "enum":
            obj = objects.Enumeration
            additional = {"base_type": self.get_type("void"), "choices": {}}
        elif type_name == "bitfield":
            obj = objects.BitField
            additional = {
                "start_bit": 0,
                "end_bit": 0,
                "base_type": self.get_type("void"),
            }
        elif type_name == "string":
            obj = objects.String
            additional = {"max_length": 0}
        elif type_name == "bytes":
            obj = objects.Bytes
            additional = {"length": 0}
        if obj is not None:
            return objects.templates.ObjectTemplate(
                obj, type_name=prefix + type_name, **additional
            )

        _native_type, native_format = self._native_dictionary[type_name]
        if type_name == "pointer":
            additional = {"subtype": self.get_type("void")}
        return objects.templates.ObjectTemplate(
            self.get_type_class(type_name),  # pylint: disable=W0142
            type_name=prefix + type_name,
            data_format=objects.DataFormatInfo(*native_format),
            **additional,
        )


std_ctypes = {
    "int": (objects.Integer, (4, "little", True)),
    "long": (objects.Integer, (4, "little", True)),
    "unsigned long": (objects.Integer, (4, "little", False)),
    "unsigned int": (objects.Integer, (4, "little", False)),
    "char": (objects.Integer, (1, "little", True)),
    "byte": (objects.Bytes, (1, "little", True)),
    "unsigned char": (objects.Integer, (1, "little", False)),
    "unsigned short int": (objects.Integer, (2, "little", False)),
    "unsigned short": (objects.Integer, (2, "little", False)),
    "unsigned be short": (objects.Integer, (2, "big", False)),
    "short": (objects.Integer, (2, "little", True)),
    "long long": (objects.Integer, (8, "little", True)),
    "unsigned long long": (objects.Integer, (8, "little", True)),
    "float": (objects.Float, (4, "little", True)),
    "double": (objects.Float, (8, "little", True)),
    "wchar": (objects.Integer, (2, "little", False)),
}
native_types = std_ctypes.copy()
native_types["pointer"] = (objects.Pointer, (4, "little", False))
x86NativeTable = NativeTable("native", native_types)
native_types["pointer"] = (objects.Pointer, (8, "little", False))
x64NativeTable = NativeTable("native", native_types)
